"""
Module to implement a plugin that ensures that blank lines surround fenced block quotes.
"""

import copy
from dataclasses import dataclass
from typing import List, Optional, Tuple, Union, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.plugins.utils.leading_space_index_tracker import (
    LeadingSpaceIndexTracker,
)
from pymarkdown.tokens.blank_line_markdown_token import BlankLineMarkdownToken
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken

# pylint: disable=too-many-lines


@dataclass(frozen=True)
class PendingContainerAdjustment:
    """
    Keep track of the adjustments we need to make on the container.
    """

    insert_index: int
    leading_space_to_insert: str
    do_insert: bool = True
    do_special: bool = False


@dataclass
class Fixer:
    """
    Keep track of fixes that need to be applied.
    """

    token: MarkdownToken
    plugin_action: str
    field_name: str
    field_value: Union[str, int]


# pylint: disable=too-many-instance-attributes
class RuleMd031(RulePlugin):
    """
    Class to implement a plugin that ensures that blank lines surround fenced block quotes.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__trigger_in_list_items: bool = True
        self.__end_fenced_code_block_token: Optional[EndMarkdownToken] = None
        self.__last_non_end_token: Optional[MarkdownToken] = None
        self.__last_token: Optional[MarkdownToken] = None
        self.__second_last_token: Optional[MarkdownToken] = None
        self.__container_adjustments: List[List[PendingContainerAdjustment]] = []
        self.__container_x: List[int] = []

        self.__fix_count = 0
        self.__removed_container_stack_token: Optional[MarkdownToken] = None
        self.__x1: List[MarkdownToken] = []
        self.__x2: List[List[PendingContainerAdjustment]] = []
        self.__x3: List[int] = []
        self.__x4: List[EndMarkdownToken] = []
        self.__x5: List[EndMarkdownToken] = []
        self.__removed_container_adjustments: Optional[
            List[PendingContainerAdjustment]
        ] = None
        self.__last_end_container_tokens: List[MarkdownToken] = []

        self.__fix_requests: List[Fixer] = []

        self.__leading_space_index_tracker = LeadingSpaceIndexTracker()

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="blanks-around-fences",
            plugin_id="MD031",
            plugin_enabled_by_default=True,
            plugin_description="Fenced code blocks should be surrounded by blank lines",
            plugin_version="0.7.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md031.md",
            plugin_configuration="list_items",
            plugin_supports_fix=True,
        )

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__trigger_in_list_items = (
            self.plugin_configuration.get_boolean_property_with_default(
                "list_items", True
            )
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("list_items", self.__trigger_in_list_items),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__last_non_end_token = None
        self.__last_token = None
        self.__end_fenced_code_block_token = None
        self.__container_adjustments = []
        self.__container_x = []
        self.__fix_count = 0
        self.__leading_space_index_tracker.clear()
        self.__removed_container_adjustments = None
        self.__removed_container_stack_token = None
        self.__x1.clear()
        self.__x2.clear()
        self.__x3.clear()
        self.__x4.clear()
        self.__x5.clear()

    def __xabc(self, context: PluginScanContext, token: MarkdownToken) -> None:
        assert (
            self.__last_token is not None
        ), "Special case means at least a block token."
        new_token = copy.deepcopy(token)
        self.__fix_count += 1
        new_token.adjust_line_number(context, self.__fix_count)
        last_token_before_code_block = self.__last_token

        # because this is the special handling, from __calculate_special_case, we know that
        # we must have 2 block quotes.
        search_index = self.__leading_space_index_tracker.get_container_stack_size() - 3
        while search_index >= 0:
            stack_token = self.__leading_space_index_tracker.get_container_stack_item(
                search_index
            )
            if stack_token.is_list_start:
                self.__container_adjustments[search_index].append(
                    PendingContainerAdjustment(0, "")
                )
                assert self.__last_token.is_end_token
                old_end_token = cast(EndMarkdownToken, self.__last_token)
                last_token_before_code_block = EndMarkdownToken(
                    old_end_token.type_name,
                    old_end_token.extracted_whitespace,
                    "",
                    old_end_token.start_markdown_token,
                    old_end_token.was_forced,
                    old_end_token.line_number,
                    old_end_token.column_number,
                )
                break
            search_index -= 1

        replacement_tokens = [
            BlankLineMarkdownToken(
                extracted_whitespace="",
                position_marker=PositionMarker(new_token.line_number - 1, 0, ""),
            ),
            last_token_before_code_block,
            new_token,
        ]
        self.register_replace_tokens_request(
            context, self.__last_token, token, replacement_tokens
        )

    def __fix_spacing_special_case(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:

        self.__xabc(context, token)
        end_token = cast(EndMarkdownToken, self.__last_token)
        block_quote_start_token = cast(
            BlockQuoteMarkdownToken, end_token.start_markdown_token
        )
        assert (
            block_quote_start_token.bleading_spaces is not None
        ), "At least one line should have been processed."
        split_bleading_spaces = block_quote_start_token.bleading_spaces.split("\n")
        self.__container_adjustments[-1].append(
            PendingContainerAdjustment(
                len(split_bleading_spaces) - 1, split_bleading_spaces[-1].rstrip()
            )
        )

    def __fix_spacing_special_case2(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        """
        This is another special case where we have to do things long-form.
        """
        assert (
            self.__last_token is not None
        ), "Special case means at least a block token."
        assert (
            self.__last_token is not None
        ), "Special case means at least a block token."
        new_token = copy.deepcopy(token)
        self.__fix_count += 1
        new_token.adjust_line_number(context, self.__fix_count)

        assert self.__second_last_token is not None
        replacement_tokens = [
            BlankLineMarkdownToken(
                extracted_whitespace="",
                position_marker=PositionMarker(new_token.line_number - 1, 0, ""),
            ),
            self.__second_last_token,
            self.__last_token,
            new_token,
        ]
        self.register_replace_tokens_request(
            context, self.__second_last_token, token, replacement_tokens
        )

        end_token = cast(EndMarkdownToken, self.__second_last_token)
        stack_token = self.__leading_space_index_tracker.get_container_stack_item(-1)
        assert stack_token == end_token.start_markdown_token
        list_start_token = cast(ListStartMarkdownToken, stack_token)
        assert (
            list_start_token.leading_spaces is not None
        ), "At least one line should have been processed."
        split_leading_spaces = list_start_token.leading_spaces.split("\n")
        self.__container_adjustments[-1].append(
            PendingContainerAdjustment(
                len(split_leading_spaces) - 1,
                split_leading_spaces[-1].rstrip(),
                do_insert=False,
            )
        )

        end_token = cast(EndMarkdownToken, self.__last_token)
        stack_token = self.__leading_space_index_tracker.get_container_stack_item(-2)
        assert stack_token == end_token.start_markdown_token
        block_quote_start_token = cast(
            BlockQuoteMarkdownToken, end_token.start_markdown_token
        )
        assert (
            block_quote_start_token.bleading_spaces is not None
        ), "At least one line should have been processed."
        split_bleading_spaces = block_quote_start_token.bleading_spaces.split("\n")
        self.__container_adjustments[-2].append(
            PendingContainerAdjustment(
                len(split_bleading_spaces) - 1,
                split_bleading_spaces[-1].rstrip(),
                do_insert=False,
            )
        )

        stack_token = self.__leading_space_index_tracker.get_container_stack_item(-3)
        list_start_token = cast(ListStartMarkdownToken, stack_token)
        assert (
            list_start_token.leading_spaces is not None
        ), "At least one line should have been processed."
        split_leading_spaces = list_start_token.leading_spaces.split("\n")
        self.__container_adjustments[-3].append(
            PendingContainerAdjustment(
                len(split_bleading_spaces) - 1, split_leading_spaces[-1]
            )
        )

        stack_token = self.__leading_space_index_tracker.get_container_stack_item(-4)
        block_quote_start_token = cast(BlockQuoteMarkdownToken, stack_token)
        assert (
            block_quote_start_token.bleading_spaces is not None
        ), "At least one line should have been processed."
        split_bleading_spaces = block_quote_start_token.bleading_spaces.split("\n")
        self.__container_adjustments[-4].append(
            PendingContainerAdjustment(
                len(split_bleading_spaces) - 1,
                split_bleading_spaces[-1],
                do_special=True,
            )
        )

    def __abc(self, context: PluginScanContext) -> Tuple[bool, int]:
        apply_rstrip = True
        delayed_container_x_adjustment = 0
        if (
            len(self.__x1) >= 3
            and self.__x1[0].is_list_start
            and self.__x1[1].is_block_quote_start
        ):
            self.__apply_tailing_block_quote_fix(1, context)
            apply_rstrip = False
            if self.__x1[2].is_block_quote_start:
                delayed_container_x_adjustment = -1
        elif (
            len(self.__x1) >= 3
            and self.__x1[0].is_list_start
            and self.__x1[1].is_list_start
            and self.__x1[2].is_block_quote_start
        ):
            self.__apply_tailing_block_quote_fix(2, context)
            apply_rstrip = False
            delayed_container_x_adjustment = -1
        elif (
            len(self.__x1) >= 3
            and self.__x1[0].is_block_quote_start
            and self.__x1[1].is_list_start
            and self.__x1[2].is_list_start
        ):
            self.__apply_tailing_block_quote_fix(0, context)
            apply_rstrip = False
            delayed_container_x_adjustment = -1
        elif len(self.__x1) >= 2 and self.__x1[0].is_block_quote_start:
            self.__apply_tailing_block_quote_fix(0, context)
        return apply_rstrip, delayed_container_x_adjustment

    def __fix_spacing_block_quote(
        self, context: PluginScanContext, token: MarkdownToken, upgrade_kludge: bool
    ) -> None:
        container_index = (
            self.__leading_space_index_tracker.get_container_stack_size() - 1
        )
        block_quote_token = cast(
            BlockQuoteMarkdownToken,
            self.__leading_space_index_tracker.get_container_stack_item(
                container_index
            ),
        )
        assert block_quote_token.bleading_spaces is not None
        split_leading_space = block_quote_token.bleading_spaces.split("\n")

        leading_space_insert_index = self.__leading_space_index_tracker.get_tokens_block_quote_bleading_space_index(
            token
        )
        former_item_leading_space = split_leading_space[leading_space_insert_index]
        apply_rstrip = True
        delayed_container_x_adjustment = 0
        if (
            leading_space_insert_index == len(split_leading_space) - 1
            and not former_item_leading_space
            and leading_space_insert_index
        ):
            former_item_leading_space = split_leading_space[
                leading_space_insert_index - 1
            ]
        elif (
            leading_space_insert_index == -1
            and len(self.__x1) >= 2
            and self.__x1[0].is_block_quote_start
            and self.__container_x
        ):
            leading_space_insert_index, upgrade_kludge, former_item_leading_space = (
                self.__fix_spacing_block_quote_1()
            )
        else:
            apply_rstrip, delayed_container_x_adjustment = self.__abc(context)

        if not upgrade_kludge and apply_rstrip:
            former_item_leading_space = former_item_leading_space.rstrip()
        self.__container_adjustments[container_index].append(
            PendingContainerAdjustment(
                leading_space_insert_index + self.__container_x[container_index],
                former_item_leading_space,
                do_special=upgrade_kludge,
            )
        )
        self.__fix_spacing_block_quote_end(token, container_index)
        self.__container_x[-1] += delayed_container_x_adjustment

    def __fix_spacing_block_quote_1(self) -> Tuple[int, bool, str]:
        end_token = self.__x4[0]
        assert end_token.start_markdown_token == self.__x1[0]
        assert end_token.extra_end_data
        self.__fix_requests.append(
            Fixer(
                end_token,
                "next_token",
                "extra_end_data",
                end_token.extra_end_data.rstrip(),
            )
        )
        leading_space_insert_index = 0
        former_item_leading_space = end_token.extra_end_data

        block_quote_token = cast(
            BlockQuoteMarkdownToken, end_token.start_markdown_token
        )
        assert block_quote_token.bleading_spaces is not None
        xxx = block_quote_token.bleading_spaces.split("\n")
        xxx[-1] = xxx[-1].rstrip()

        self.__fix_requests.append(
            Fixer(
                end_token.start_markdown_token,
                "next_token",
                "bleading_spaces",
                "\n".join(xxx),
            )
        )

        upgrade_kludge = True
        self.__container_x[-1] += 1
        return leading_space_insert_index, upgrade_kludge, former_item_leading_space

    def __fix_spacing_block_quote_end(
        self, token: MarkdownToken, container_index: int
    ) -> None:

        while (
            container_index > 0
            and not self.__leading_space_index_tracker.get_container_stack_item(
                container_index - 1
            ).is_list_start
        ):
            container_index -= 1

        if (
            container_index > 0
            and self.__leading_space_index_tracker.get_container_stack_item(
                container_index - 1
            ).is_list_start
        ):
            leading_space_insert_index = (
                LeadingSpaceIndexTracker.calculate_token_line_number(token)
                - self.__leading_space_index_tracker.get_container_stack_item(
                    container_index - 1
                ).line_number
            )
            self.__container_adjustments[container_index - 1].append(
                PendingContainerAdjustment(leading_space_insert_index, "")
            )

    def __fix_spacing_list_special(
        self,
        context: PluginScanContext,
        list_token_mod_index: int,
        previous_block_quote_index: int,
    ) -> None:
        _ = context
        end_adj = self.__removed_container_adjustments
        assert not end_adj

        outer_block_quote_token = cast(
            BlockQuoteMarkdownToken, self.__removed_container_stack_token
        )
        assert outer_block_quote_token.bleading_spaces is not None
        assert (
            outer_block_quote_token.bleading_spaces[-1] == " "
            and outer_block_quote_token.bleading_spaces[-2] == ">"
        )
        stripped_bleading_spaces = outer_block_quote_token.bleading_spaces[:-1]
        self.__fix_requests.append(
            Fixer(
                outer_block_quote_token,
                "next_token",
                "bleading_spaces",
                stripped_bleading_spaces,
            )
        )

        current_list_token = (
            self.__leading_space_index_tracker.get_container_stack_item(-1)
        )
        inner_block_quote_token = (
            self.__leading_space_index_tracker.get_container_stack_item(
                previous_block_quote_index
            )
        )
        adjusted_indent_level = cast(
            ListStartMarkdownToken, current_list_token
        ).indent_level - (inner_block_quote_token.column_number + 1)

        self.__container_adjustments[-1].append(
            PendingContainerAdjustment(list_token_mod_index, "")
        )
        self.__container_adjustments[-1].append(
            PendingContainerAdjustment(
                list_token_mod_index, " " * adjusted_indent_level, do_insert=False
            )
        )

        leading_space_to_insert = "> "
        remaining_index = previous_block_quote_index - 1
        if remaining_index >= 0:
            assert remaining_index == 0
            container_stack_token = (
                self.__leading_space_index_tracker.get_container_stack_item(
                    remaining_index
                )
            )
            if container_stack_token.is_list_start:
                list_start_token = cast(ListStartMarkdownToken, container_stack_token)
                leading_space_to_insert = (
                    " " * list_start_token.indent_level
                ) + leading_space_to_insert
            else:
                # block_quote_token = cast(BlockQuoteMarkdownToken, container_stack_token)
                leading_space_to_insert = f"> {leading_space_to_insert}"
        block_quote_mod_index = (
            outer_block_quote_token.line_number
            - inner_block_quote_token.line_number
            + 1
        )
        self.__container_adjustments[previous_block_quote_index].append(
            PendingContainerAdjustment(block_quote_mod_index, leading_space_to_insert)
        )

    def __fix_spacing_list_detect_special(
        self, token: MarkdownToken
    ) -> Tuple[bool, int, int]:
        if (
            self.__removed_container_stack_token
            and self.__removed_container_stack_token.is_block_quote_start
        ):

            initial_index = (
                self.__leading_space_index_tracker.get_container_stack_size() - 1
            )

            list_token = self.__leading_space_index_tracker.get_container_stack_item(
                initial_index
            )
            assert list_token.is_list_start
            leading_spaces = cast(ListStartMarkdownToken, list_token).leading_spaces
            assert leading_spaces is not None
            split_spaces = leading_spaces.split("\n")

            # last_closed_container_info = (
            #     self.__leading_space_index_tracker.get_closed_container_info(-1)
            # )

            # if last_closed_container_info.adjustment:
            adjust = 1
            # else:
            #     adjust = self.__calculate_adjust(initial_index, container_index)
            index = (
                LeadingSpaceIndexTracker.calculate_token_line_number(token)
                - self.__leading_space_index_tracker.get_container_stack_item(
                    initial_index
                ).line_number
            )
            # index -= last_closed_container_info.adjustment
            index -= adjust
            # SNAFU6?
            xx = self.__leading_space_index_tracker.get_tokens_list_leading_space_index(
                token
            )
            selected_leading_space = split_spaces[xx]
            if selected_leading_space.endswith(ParserLogger.blah_sequence):
                stack_index = initial_index - 1
                while (
                    stack_index >= 0
                    and not self.__leading_space_index_tracker.get_container_stack_item(
                        stack_index
                    ).is_block_quote_start
                ):
                    stack_index -= 1
                assert stack_index >= 0
                return True, stack_index, xx
        return False, -1, -1

    def __fix_spacing_list(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> Tuple[bool, bool]:
        is_special, stack_index, index = self.__fix_spacing_list_detect_special(token)
        if is_special:
            self.__fix_spacing_list_special(context, index, stack_index)
            return True, False

        initial_index = container_index = (
            self.__leading_space_index_tracker.get_container_stack_size() - 1
        )
        while (
            container_index >= 0
            and self.__leading_space_index_tracker.get_container_stack_item(
                container_index - 1
            ).is_list_start
        ):
            container_index -= 1
        if container_index >= 0:

            block_quote_index, index, _ = self.__fix_spacing_list_prefix(
                context, token, container_index, initial_index
            )

            assert block_quote_index.bleading_spaces is not None
            split_bleading_spaces = block_quote_index.bleading_spaces.split("\n")
            self.__container_adjustments[container_index - 1].append(
                PendingContainerAdjustment(index, split_bleading_spaces[index].rstrip())
            )
            # this may be due to a commented out test
            # Above, __fix_spacing_list_prefix has a third return value of ss
            # assert ss is None
            # self.__container_adjustments[container_index - 1].append(
            #     PendingContainerAdjustment(index, ss, do_insert=False)
            # )

        kludge_alpha = False
        if (
            self.__removed_container_stack_token is not None
            and not self.__removed_container_stack_token.is_block_quote_start
            # and container_index
        ):
            kludge_alpha = self.__fix_spacing_list_remove_list(context, token)
        else:
            self.__fix_spacing_list_not_remove_list(
                initial_index, container_index, token
            )
        return False, kludge_alpha

    def __xxabc(self, split_spaces: List[str]) -> bool:
        xxxxx = split_spaces.pop(-1)
        split_spaces.append("")
        search_index = 0
        self.__container_adjustments[search_index].append(
            PendingContainerAdjustment(0, xxxxx)
        )
        return True

    # pylint: disable=too-many-boolean-expressions
    def __fix_spacing_list_remove_list_kludge(self, token: MarkdownToken) -> bool:
        if (
            len(self.__x1) == 2
            and self.__x1[0].is_list_start
            and self.__x1[1].is_list_start
            and self.__leading_space_index_tracker.get_container_stack_size() == 2
            and self.__leading_space_index_tracker.get_container_stack_item(
                0
            ).is_block_quote_start
            and self.__leading_space_index_tracker.get_container_stack_item(
                1
            ).is_list_start
        ):
            list_token = cast(ListStartMarkdownToken, self.__x1[0])
            assert list_token.leading_spaces is not None
            split_leading_spaces = list_token.leading_spaces.split("\n")
            split_leading_spaces.insert(self.__x3[0] - 1, "")
            self.__leading_space_index_tracker.nudge_list_container(
                token, nudge_positive=False
            )
            self.__fix_requests.append(
                Fixer(
                    self.__x1[0],
                    "next_token",
                    "leading_spaces",
                    "\n".join(split_leading_spaces),
                )
            )
            return False
        return True

    # pylint: enable=too-many-boolean-expressions

    def __fix_spacing_list_remove_list(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> bool:
        _ = context
        if not self.__fix_spacing_list_remove_list_kludge(token):
            return False

        xx = self.__removed_container_stack_token
        removed_list_token = cast(ListStartMarkdownToken, xx)
        if removed_list_token.leading_spaces is None:
            return False

        kludge_alpha = False
        split_spaces = removed_list_token.leading_spaces.split("\n")
        split_spaces_length = len(split_spaces)
        if split_spaces_length > 1:

            if len(self.__x3) > 1 and self.__x3[-1] < split_spaces_length:
                assert self.__x3[-1] == split_spaces_length - 1
                kludge_alpha = self.__xxabc(split_spaces)
            else:
                split_spaces.insert(split_spaces_length - 1, "")
        else:
            if (
                len(self.__x1) == 3
                and self.__x1[0].is_list_start
                and self.__x1[1].is_block_quote_start
                and self.__x1[2].is_list_start
            ):
                assert self.__x3[-1] == split_spaces_length
                kludge_alpha = self.__xxabc(split_spaces)
            split_spaces.append("")
        assert removed_list_token is not None
        self.__fix_requests.append(
            Fixer(
                removed_list_token,
                "next_token",
                "leading_spaces",
                "\n".join(split_spaces),
            )
        )
        return kludge_alpha

    def __xx(self, container_index: int, initial_index: int) -> Tuple[bool, int, int]:
        continue_processing = True
        adjust = 0
        adj2 = 0
        if container_index == -1:
            if (
                len(self.__x1) == 2
                and self.__x1[0].is_list_start
                and self.__x1[1].is_block_quote_start
            ):
                adjust = 1
                adj2 = 1
                continue_processing = False
            elif (
                len(self.__x1) == 3
                and self.__x1[0].is_list_start
                and self.__x1[1].is_block_quote_start
                and self.__x1[2].is_block_quote_start
            ):
                adjust = 1
                adj2 = 1
                continue_processing = False
            elif (
                len(self.__x1) == 3
                and self.__x1[0].is_block_quote_start
                and self.__x1[1].is_list_start
                and self.__x1[2].is_block_quote_start
            ):
                self.__xx1(initial_index)
                continue_processing = False
                adjust = 1
        return continue_processing, adjust, adj2

    def __xx1(self, initial_index: int) -> None:
        gh = cast(ListStartMarkdownToken, self.__x1[1]).leading_spaces
        assert gh is not None
        split_spaces = gh.split("\n")
        assert self.__x3[1] == len(split_spaces) - 1
        sdf = split_spaces[-1]
        del split_spaces[-1]
        # split_spaces.append("")
        self.__fix_requests.append(
            Fixer(
                self.__x1[1],
                "next_token",
                "leading_spaces",
                "\n".join(split_spaces),
            )
        )
        self.__container_adjustments[initial_index].append(
            PendingContainerAdjustment(0, sdf, do_special=True)
        )

    def __fix_spacing_list_not_remove_list(
        self, initial_index: int, container_index: int, token: MarkdownToken
    ) -> None:
        last_closed_container_info = (
            self.__leading_space_index_tracker.get_closed_container_info(-1)
        )

        continue_processing, adjust, adj2 = self.__xx(container_index, initial_index)
        if continue_processing:
            if last_closed_container_info.adjustment:
                adjust = 2 if container_index >= 0 else 1
            else:
                adjust = self.__calculate_adjust(initial_index, container_index)
        if not continue_processing:
            last_closed_container_info.adjustment += adj2
        index = (
            LeadingSpaceIndexTracker.calculate_token_line_number(token)
            - self.__leading_space_index_tracker.get_container_stack_item(
                initial_index
            ).line_number
        )
        index -= (
            last_closed_container_info.adjustment
            + last_closed_container_info.list_nudge_count
        )
        index -= adjust
        if index < 0:
            index = 0
            self.__leading_space_index_tracker.nudge_list_container(
                token, nudge_positive=False
            )
        self.__container_adjustments[initial_index].append(
            PendingContainerAdjustment(index, "")
        )

    def __fix_spacing_list_prefix(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        container_index: int,
        initial_index: int,
    ) -> Tuple[BlockQuoteMarkdownToken, int, Optional[str]]:
        block_quote_index = cast(
            BlockQuoteMarkdownToken,
            self.__leading_space_index_tracker.get_container_stack_item(
                container_index - 1
            ),
        )

        current_closed_container_info = (
            self.__leading_space_index_tracker.get_closed_container_info(
                container_index - 1
            )
        )

        index = (
            LeadingSpaceIndexTracker.calculate_token_line_number(token)
            - block_quote_index.line_number
            - current_closed_container_info.adjustment
        )
        if current_closed_container_info.adjustment != 0:
            index += current_closed_container_info.count

        # ss = None
        # This may be due to a commented out test.
        #             # SNAFU6?

        # IF
        assert not (
            container_index == initial_index
            and self.__last_token is not None
            and self.__last_token.is_block_quote_end
        )
        _ = context
        ss = None
        # if (
        #     container_index == initial_index
        #     and self.__last_token is not None
        #     and self.__last_token.is_block_quote_end
        # ):
        #     x = cast(EndMarkdownToken, self.__last_token)
        #     if x.extra_end_data is not None:
        #         ss = x.extra_end_data
        #         self.register_fix_token_request(
        #             context, x, "next_token", "extra_end_data", ""
        #         )
        #         self.__container_adjustments[container_index - 1].append(
        #             PendingContainerAdjustment(index, ss)
        #         )
        return block_quote_index, index, ss

    def __calculate_adjust(self, initial_index: int, container_index: int) -> int:

        last_closed_container_info = (
            self.__leading_space_index_tracker.get_closed_container_info(-1)
        )
        assert (
            initial_index < 2
            or container_index
            or not last_closed_container_info.adjustment
        )
        return (
            0
            if initial_index >= 1
            and not container_index
            and last_closed_container_info.adjustment
            else 1
        )

    def __fix_spacing_one_container(
        self, context: PluginScanContext, token: MarkdownToken, upgrade_kludge: bool
    ) -> Tuple[bool, bool]:
        did_special_list_fix = False
        kludge_alpha = False
        if self.__leading_space_index_tracker.get_container_stack_item(
            -1
        ).is_block_quote_start:
            self.__fix_spacing_block_quote(context, token, upgrade_kludge)
        else:
            did_special_list_fix, kludge_alpha = self.__fix_spacing_list(context, token)
        return did_special_list_fix, kludge_alpha

    def __fix_spacing_removed_container(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        _ = (token, context)
        assert self.__removed_container_stack_token is not None
        if self.__removed_container_stack_token.is_list_start:
            removed_list_token = cast(
                ListStartMarkdownToken, self.__removed_container_stack_token
            )
            if removed_list_token.leading_spaces is not None:
                split_spaces = removed_list_token.leading_spaces.split("\n")
                split_spaces.append("")
            else:
                split_spaces = [""]
            self.__fix_requests.append(
                Fixer(
                    self.__removed_container_stack_token,
                    "next_token",
                    "leading_spaces",
                    "\n".join(split_spaces),
                )
            )

    def __fix_spacing_with_fenced_and_list_end(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        new_token: MarkdownToken,
        kludge_alpha: bool,
    ) -> None:
        end_container_index = len(self.__last_end_container_tokens) - 1
        while (
            end_container_index >= 0
            and self.__last_end_container_tokens[end_container_index].is_list_end
        ):
            end_container_index -= 1
        first_token = self.__last_end_container_tokens[end_container_index + 1]
        replacement_tokens: List[MarkdownToken] = []
        if end_container_index >= 0:
            if not kludge_alpha:
                replacement_tokens.extend(
                    self.__last_end_container_tokens[end_container_index + 1 :]
                )
            replacement_tokens.append(
                BlankLineMarkdownToken(
                    extracted_whitespace="",
                    position_marker=PositionMarker(new_token.line_number - 1, 0, ""),
                    column_delta=1,
                )
            )
            if kludge_alpha:
                replacement_tokens.extend(
                    self.__last_end_container_tokens[end_container_index + 1 :]
                )
        else:
            replacement_tokens.append(
                BlankLineMarkdownToken(
                    extracted_whitespace="",
                    position_marker=PositionMarker(new_token.line_number - 1, 0, ""),
                    column_delta=1,
                )
            )
            replacement_tokens.extend(
                self.__last_end_container_tokens[end_container_index + 1 :]
            )
        replacement_tokens.append(new_token)
        self.register_replace_tokens_request(
            context, first_token, token, replacement_tokens
        )

    def __fix_spacing_with_special_list_fix(
        self, context: PluginScanContext, token: MarkdownToken, new_token: MarkdownToken
    ) -> None:
        assert self.__last_token and self.__last_token.is_block_quote_end
        new_end_token = cast(EndMarkdownToken, copy.copy(self.__last_token))
        new_end_token.set_extra_end_data(None)
        replacement_tokens = [
            BlankLineMarkdownToken(
                extracted_whitespace="",
                position_marker=PositionMarker(new_token.line_number - 1, 0, ""),
                column_delta=1,
            ),
            new_end_token,
            new_token,
        ]
        self.register_replace_tokens_request(
            context, self.__last_token, token, replacement_tokens
        )

    def __fix_spacing_with_else(
        self, context: PluginScanContext, token: MarkdownToken, new_token: MarkdownToken
    ) -> None:
        replacement_tokens = [
            BlankLineMarkdownToken(
                extracted_whitespace="",
                position_marker=PositionMarker(new_token.line_number - 1, 0, ""),
                column_delta=1,
            ),
            new_token,
        ]
        self.register_replace_tokens_request(context, token, token, replacement_tokens)

    def __calc_kludge_one(self, at_least_one_container: bool) -> bool:
        if not at_least_one_container:
            return False
        last_stack_token = self.__leading_space_index_tracker.get_container_stack_item(
            -1
        )
        return (
            not any(i.is_block_quote_start for i in self.__x1[:-1])
            if last_stack_token.is_list_start and self.__x1[-1].is_block_quote_start
            else False
        )

    def __calc_3(
        self,
        context: PluginScanContext,
        did_process_removals: bool,
        at_least_one_container: bool,
        upgrade_kludge: bool,
    ) -> Tuple[bool, bool]:
        # This will most likely need rewriting for deeper nestings.
        if not did_process_removals and at_least_one_container:
            last_stack_token_index = (
                self.__leading_space_index_tracker.get_container_stack_size() - 1
            )
            found_block_quote_token = None
            assert last_stack_token_index >= 0
            token_at_index = (
                self.__leading_space_index_tracker.get_container_stack_item(
                    last_stack_token_index
                )
            )
            if token_at_index.is_block_quote_start:  # pragma: no cover
                found_block_quote_token = token_at_index
                while last_stack_token_index >= 0:
                    if (
                        token_at_index := self.__leading_space_index_tracker.get_container_stack_item(
                            last_stack_token_index
                        )
                    ).is_block_quote_start:
                        found_block_quote_token = token_at_index
                        break
                    last_stack_token_index -= 1

                if (
                    found_block_quote_token
                    and len(self.__x1) == 2
                    and self.__x1[0].is_list_start
                    and self.__x1[1].is_block_quote_start
                ):
                    # dd = 0
                    # while dd < len(self.__x1) and not self.__x1[dd].is_block_quote_start:
                    #     dd += 1
                    # assert dd < len(self.__x1)
                    dd = 1
                    did_process_removals = upgrade_kludge = True
                    self.__apply_tailing_block_quote_fix(dd, context)
        return did_process_removals, upgrade_kludge

    def __apply_tailing_block_quote_fix(
        self, modify_index: int, context: PluginScanContext
    ) -> None:
        _ = context

        assert self.__x1[modify_index].is_block_quote_start
        block_quote_token = cast(BlockQuoteMarkdownToken, self.__x1[modify_index])
        assert block_quote_token.bleading_spaces is not None
        split_spaces = block_quote_token.bleading_spaces.split("\n")
        assert self.__x3[modify_index] == len(split_spaces) - 1
        split_spaces[-1] = split_spaces[-1].rstrip()
        self.__fix_requests.append(
            Fixer(
                self.__x1[modify_index],
                "next_token",
                "bleading_spaces",
                "\n".join(split_spaces),
            )
        )

    def __fix_spacing(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        special_case: bool,
        special_case_2: bool,
    ) -> None:
        if special_case:
            self.__fix_spacing_special_case(context, token)
            return
        if special_case_2:
            self.__fix_spacing_special_case2(context, token)
            return
        at_least_one_container = (
            self.__leading_space_index_tracker.in_at_least_one_container()
        )
        upgrade_kludge = False
        if len(self.__x1) > 1:
            # These circumstances are already handled by the zero and single level removals.
            # all_removals_are_block_quotes = all(
            #     f.is_block_quote_start for f in self.__x1
            # )
            all_removals_are_lists = all(f.is_list_start for f in self.__x1)
            was_kludge_one = self.__calc_kludge_one(at_least_one_container)
            did_process_removals = (
                # all_removals_are_block_quotes or
                all_removals_are_lists
                or was_kludge_one
                or not at_least_one_container
            )

            # did_process_removals = self.__calc_2(context, did_process_removals)
            did_process_removals, upgrade_kludge = self.__calc_3(
                context, did_process_removals, at_least_one_container, upgrade_kludge
            )
            # assert did_process_removals

        did_special_list_fix = False
        kludge_alpha = False
        if at_least_one_container:
            did_special_list_fix, kludge_alpha = self.__fix_spacing_one_container(
                context, token, upgrade_kludge
            )
        elif self.__removed_container_stack_token:
            self.__fix_spacing_removed_container(context, token)

        new_token = copy.deepcopy(token)
        self.__fix_count += 1
        new_token.adjust_line_number(context, self.__fix_count)
        assert self.__last_token is not None
        if token.is_fenced_code_block and self.__last_token.is_list_end:
            self.__fix_spacing_with_fenced_and_list_end(
                context, token, new_token, kludge_alpha
            )
        elif did_special_list_fix:
            self.__fix_spacing_with_special_list_fix(context, token, new_token)
        else:
            self.__fix_spacing_with_else(context, token, new_token)

    def __handle_fenced_code_block(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        special_case: bool,
        special_case_2: bool,
    ) -> None:

        can_trigger = (
            self.__trigger_in_list_items
            if self.__leading_space_index_tracker.in_at_least_one_container()
            and self.__leading_space_index_tracker.get_container_stack_item(
                -1
            ).is_list_start
            else True
        )
        if (
            self.__last_non_end_token
            and not self.__last_non_end_token.is_blank_line
            and can_trigger
        ):
            if context.in_fix_mode and not context.is_during_line_pass:
                self.__fix_spacing(context, token, special_case, special_case_2)
            else:
                self.report_next_token_error(context, token)

    def __calculate_end_deltas(self) -> Tuple[int, int]:
        line_number_delta = 0
        assert self.__last_non_end_token is not None
        if self.__last_non_end_token.is_text:
            text_token = cast(TextMarkdownToken, self.__last_non_end_token)
            line_number_delta = (
                text_token.token_text.count(ParserHelper.newline_character) + 2
            )
        else:
            assert self.__last_non_end_token.is_fenced_code_block
            line_number_delta = 1

        assert self.__end_fenced_code_block_token is not None
        column_number_delta = (
            self.__end_fenced_code_block_token.start_markdown_token.column_number
        )
        start_token = cast(
            EndMarkdownToken,
            self.__end_fenced_code_block_token.start_markdown_token,
        )
        if start_token.extracted_whitespace:
            column_number_delta -= len(start_token.extracted_whitespace)
        if (
            self.__end_fenced_code_block_token
            and self.__end_fenced_code_block_token.extracted_whitespace
        ):
            column_number_delta += len(
                self.__end_fenced_code_block_token.extracted_whitespace
            )
        column_number_delta = -(column_number_delta)

        return line_number_delta, column_number_delta

    def __handle_end_fenced_code_block(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:  # sourcery skip: extract-method
        can_trigger = not token.is_end_of_stream
        if (
            self.__leading_space_index_tracker.in_at_least_one_container()
            and self.__leading_space_index_tracker.get_container_stack_item(
                -1
            ).is_list_start
        ):
            can_trigger = self.__trigger_in_list_items
        if (
            not token.is_blank_line
            and self.__end_fenced_code_block_token is not None
            and not self.__end_fenced_code_block_token.was_forced
            and can_trigger
        ):
            if context.in_fix_mode:
                self.__fix_spacing(context, token, False, False)
            else:
                assert self.__last_non_end_token
                if self.__last_non_end_token.is_text:
                    text_token = cast(TextMarkdownToken, self.__last_non_end_token)
                    override_is_error_token_prefaced_by_blank_line = (
                        text_token.token_text.endswith(
                            ParserHelper.newline_character
                            + ParserHelper.replace_noop_character
                        )
                    )
                else:
                    override_is_error_token_prefaced_by_blank_line = False
                line_number_delta, column_number_delta = self.__calculate_end_deltas()
                error_token = self.__end_fenced_code_block_token.start_markdown_token
                self.report_next_token_error(
                    context,
                    error_token,
                    line_number_delta=line_number_delta
                    + context.calc_pragma_offset(error_token, line_number_delta),
                    column_number_delta=column_number_delta,
                    override_is_error_token_prefaced_by_blank_line=override_is_error_token_prefaced_by_blank_line,
                )
        self.__end_fenced_code_block_token = None

    def __process_pending_container_end_adjustment(
        self,
        context: PluginScanContext,
        next_container_adjustment_list: List[PendingContainerAdjustment],
    ) -> None:
        _ = context
        if self.__leading_space_index_tracker.get_container_stack_item(
            -1
        ).is_block_quote_start:
            token_part_name = "bleading_spaces"
            block_quote_token = cast(
                BlockQuoteMarkdownToken,
                self.__leading_space_index_tracker.get_container_stack_item(-1),
            )
            assert (
                block_quote_token.bleading_spaces is not None
            ), "Pending containers means this should at least have a newline in it."
            split_spaces = block_quote_token.bleading_spaces.split("\n")
        else:
            token_part_name = "leading_spaces"
            list_token = cast(
                ListStartMarkdownToken,
                self.__leading_space_index_tracker.get_container_stack_item(-1),
            )
            assert (
                list_token.leading_spaces is not None
            ), "Pending containers means this should at least have a newline in it."
            split_spaces = list_token.leading_spaces.split("\n")

        for i, next_container_adjustment in enumerate(next_container_adjustment_list):
            if next_container_adjustment.do_special:
                split_spaces.insert(
                    0, next_container_adjustment.leading_space_to_insert
                )
                del next_container_adjustment_list[i]
                break

        for next_container_adjustment in next_container_adjustment_list[::-1]:
            assert not next_container_adjustment.do_special
            if next_container_adjustment.do_insert:
                split_spaces.insert(
                    next_container_adjustment.insert_index,
                    next_container_adjustment.leading_space_to_insert,
                )
            else:
                split_spaces[next_container_adjustment.insert_index] = (
                    next_container_adjustment.leading_space_to_insert
                )

        self.__fix_requests.append(
            Fixer(
                self.__leading_space_index_tracker.get_container_stack_item(-1),
                "next_token",
                token_part_name,
                "\n".join(split_spaces),
            )
        )

    def __process_pending_container_end_tokens(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        # is_first = True
        while self.__leading_space_index_tracker.have_any_registered_container_ends():
            if context.in_fix_mode:
                if next_container_adjustment_list := self.__container_adjustments[-1]:
                    self.__process_pending_container_end_adjustment(
                        context, next_container_adjustment_list
                    )

            if self.__leading_space_index_tracker.get_container_stack_item(
                -1
            ).is_block_quote_start:
                leading_space_insert_index = self.__leading_space_index_tracker.get_tokens_block_quote_bleading_space_index(
                    token
                )
            else:
                leading_space_insert_index = self.__leading_space_index_tracker.get_tokens_list_leading_space_index(
                    token
                )
            self.__x3.append(leading_space_insert_index)

            # xxxxxxx = self.__leading_space_index_tracker.since_last_non_end_token()
            # fgg = (
            #     is_first
            #     and len(xxxxxxx) >= 3
            #     and xxxxxxx[-1].is_list_end
            #     and xxxxxxx[-2].is_block_quote_end
            #     and xxxxxxx[-3].is_list_end
            # )

            self.__removed_container_stack_token, xx = (
                self.__leading_space_index_tracker.process_container_end(token)
            )
            assert xx.start_markdown_token == self.__removed_container_stack_token
            self.__x1.append(self.__removed_container_stack_token)
            self.__x4.append(xx)

            self.__removed_container_adjustments = self.__container_adjustments[-1]
            self.__x2.append(self.__removed_container_adjustments)
            del self.__container_adjustments[-1]
            del self.__container_x[-1]
            # is_first = False

    def __calculate_special_case(self) -> bool:
        return bool(
            self.__leading_space_index_tracker.get_container_stack_size() >= 2
            and self.__leading_space_index_tracker.get_container_stack_item(
                -1
            ).is_block_quote_start
            and self.__leading_space_index_tracker.get_container_stack_item(
                -2
            ).is_block_quote_start
            and self.__last_token
            and self.__second_last_token
            and self.__last_token.is_block_quote_end
            and self.__second_last_token.is_paragraph_end
        )

    def __calculate_specials(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> Tuple[bool, bool]:

        special_case = False
        special_case_2 = False
        if context.in_fix_mode and token.is_fenced_code_block:
            special_case = self.__calculate_special_case()
            if not special_case:
                special_case_2 = (
                    len(self.__x5) >= 3
                    and self.__x5[-1].is_block_quote_end
                    and self.__x5[-2].is_list_end
                    and self.__x5[-3].is_paragraph_end
                )
                if special_case_2:
                    special_case_2 = (
                        self.__leading_space_index_tracker.get_container_stack_size()
                        >= 4
                        and self.__leading_space_index_tracker.get_container_stack_item(
                            -3
                        ).is_list_start
                        and self.__leading_space_index_tracker.get_container_stack_item(
                            -4
                        ).is_block_quote_start
                    )

        return special_case, special_case_2

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """

        special_case, special_case_2 = self.__calculate_specials(context, token)
        if not token.is_end_token or token.is_end_of_stream:
            if not (special_case or special_case_2):
                self.__process_pending_container_end_tokens(context, token)
            if self.__end_fenced_code_block_token:
                self.__handle_end_fenced_code_block(context, token)

        if token.is_block_quote_start or token.is_list_start:
            self.__container_adjustments.append([])
            self.__container_x.append(0)
            self.__leading_space_index_tracker.open_container(token)
        elif token.is_block_quote_end or token.is_list_end:
            self.__leading_space_index_tracker.register_container_end(token)
        elif token.is_fenced_code_block:
            self.__handle_fenced_code_block(
                context, token, special_case, special_case_2
            )
            if special_case or special_case_2:
                self.__process_pending_container_end_tokens(context, token)
        elif token.is_fenced_code_block_end:
            self.__end_fenced_code_block_token = cast(EndMarkdownToken, token)

        if token.is_end_token:
            self.__x5.append(cast(EndMarkdownToken, token))
        else:
            self.__x5.clear()
        self.__leading_space_index_tracker.track_since_last_non_end_token(token)

        self.__next_token_after(context, token)

    def __next_token_after(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:

        if (
            not token.is_end_token
            and not token.is_block_quote_start
            and not token.is_list_start
        ):
            self.__last_non_end_token = token

        if token.is_block_quote_end or token.is_list_end:
            self.__last_end_container_tokens.append(token)
        else:
            self.__last_end_container_tokens.clear()

        self.__second_last_token = self.__last_token
        self.__last_token = token
        self.__removed_container_stack_token = None
        self.__x1.clear()
        self.__x4.clear()
        self.__removed_container_adjustments = None
        self.__x2.clear()
        self.__x3.clear()

        while self.__fix_requests:
            ff = self.__fix_requests[0]
            del self.__fix_requests[0]
            self.register_fix_token_request(
                context, ff.token, ff.plugin_action, ff.field_name, ff.field_value
            )


# pylint: enable=too-many-instance-attributes



"""
Module to implement a plugin that looks for headings that do not start at the
beginning of the line.
"""

from typing import Dict, List, Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import PluginDetailsV2
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.plugins.utils.container_token_manager import ContainerTokenManager
from pymarkdown.tokens.atx_heading_markdown_token import AtxHeadingMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.setext_heading_markdown_token import SetextHeadingMarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class RuleMd023(RulePlugin):
    """
    Class to implement a plugin that looks for headings that do not start at the
    beginning of the line.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__setext_start_token: Optional[SetextHeadingMarkdownToken] = None
        self.__any_leading_whitespace_detected = False
        self.__seen_first_line_of_setext = False
        self.__container_manager = ContainerTokenManager()
        self.__last_skipped_text_token: Optional[TextMarkdownToken] = None
        self.__leading_spaces_split: Dict[ListStartMarkdownToken, List[str]] = {}

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV2(
            plugin_name="heading-start-left, header-start-left",
            plugin_id="MD023",
            plugin_enabled_by_default=True,
            plugin_description="Headings must start at the beginning of the line.",
            plugin_version="0.5.3",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md023.md",
            plugin_supports_fix=True,
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__setext_start_token = None
        self.__any_leading_whitespace_detected = False
        self.__seen_first_line_of_setext = False
        self.__container_manager.clear()
        self.__leading_spaces_split = {}
        self.__last_skipped_text_token = None

    def __fix_adjustments(self, ex_ws: str, ind: int = 0) -> str:
        if self.__container_manager.container_token_stack:
            stack_length = len(self.__container_manager.container_token_stack)
            atx_container_token = self.__container_manager.container_token_stack[
                stack_length - 1
            ]
        else:
            atx_container_token = None

        fixed_whitespace = ""
        if atx_container_token is not None:
            if atx_container_token.is_block_quote_start:
                fixed_whitespace = " " if ex_ws[0] == "\t" else ""
            else:
                fixed_whitespace = ""
                if ex_ws[0] == "\t":

                    list_start_token = cast(ListStartMarkdownToken, atx_container_token)
                    track_line_index = self.__container_manager.bq_line_index[
                        stack_length
                    ]
                    adjust_line_index = self.__container_manager.list_adjust_map[
                        stack_length
                    ]

                    if list_start_token in self.__leading_spaces_split:
                        split_spaces = self.__leading_spaces_split[list_start_token]
                    else:
                        assert list_start_token.leading_spaces is not None
                        split_spaces = list_start_token.leading_spaces.split("\n")
                        self.__leading_spaces_split[list_start_token] = split_spaces
                    split_spaces[track_line_index - adjust_line_index + ind] = (
                        " " * list_start_token.indent_level
                    )
        return fixed_whitespace

    def __handle_atx_heading(
        self, context: PluginScanContext, token: AtxHeadingMarkdownToken
    ) -> None:
        if not token.extracted_whitespace:
            return

        if context.in_fix_mode:
            self.register_fix_token_request(
                context,
                token,
                "next_token",
                "extracted_whitespace",
                self.__fix_adjustments(token.extracted_whitespace),
            )
        else:
            self.report_next_token_error(context, token)

    def __handle_setext_heading(
        self, context: PluginScanContext, token: SetextHeadingMarkdownToken
    ) -> None:
        self.__setext_start_token = token
        self.__any_leading_whitespace_detected = bool(token.extracted_whitespace)
        if self.__any_leading_whitespace_detected and context.in_fix_mode:
            assert token.extracted_whitespace
            whitespace_to_add = " " if token.extracted_whitespace[0] == "\t" else ""
            self.register_fix_token_request(
                context,
                token,
                "next_token",
                "extracted_whitespace",
                whitespace_to_add,
            )
        self.__seen_first_line_of_setext = False
        self.__last_skipped_text_token = None

    def __handle_setext_heading_end(
        self, context: PluginScanContext, token: EndMarkdownToken
    ) -> None:
        if self.__last_skipped_text_token and context.in_fix_mode:
            self.__handle_text(context, self.__last_skipped_text_token, True)

        if token.extracted_whitespace:
            self.__any_leading_whitespace_detected = True
            if context.in_fix_mode:
                self.register_fix_token_request(
                    context,
                    token,
                    "next_token",
                    "extracted_whitespace",
                    self.__fix_adjustments(token.extracted_whitespace),
                )

        if self.__any_leading_whitespace_detected:
            assert self.__setext_start_token is not None
            self.report_next_token_error(
                context, self.__setext_start_token, use_original_position=True
            )
        self.__setext_start_token = None

    def __handle_text_split_end(
        self, next_split_end_whitespace: str, split_index: int, new_end_parts: List[str]
    ) -> None:
        split_next_split = next_split_end_whitespace.split(
            ParserHelper.whitespace_split_character
        )
        if len(split_next_split) == 2 and split_next_split[0]:
            self.__any_leading_whitespace_detected = True
            fix_string = self.__fix_adjustments(split_next_split[0], ind=split_index)
            new_split_value = (
                fix_string
                + ParserHelper.whitespace_split_character
                + split_next_split[1]
            )
            if new_split_value == ParserHelper.whitespace_split_character:
                new_end_parts.append("")
            else:
                new_end_parts.append(new_split_value)
        elif next_split_end_whitespace:
            new_end_parts.append(
                ParserHelper.whitespace_split_character + next_split_end_whitespace
            )
        else:
            new_end_parts.append(next_split_end_whitespace)

    # pylint: disable=too-many-arguments
    def __handle_text_split(
        self,
        context: PluginScanContext,
        split_text: List[str],
        split_end_whitespace: Optional[List[str]],
        split_index: int,
        new_text_parts: List[str],
        new_end_parts: List[str],
        is_setext_end: bool,
    ) -> None:
        next_split_end_whitespace = None
        if split_end_whitespace is not None:
            next_split_end_whitespace = split_end_whitespace[split_index]
        next_split_text = split_text[split_index]

        if context.in_fix_mode:
            if next_split_text and ParserHelper.is_character_replacement_marker(
                next_split_text, 0
            ):
                start_index, _, end_index = ParserHelper.get_replacement_indices(
                    next_split_text, 0
                )
                assert start_index == 0
                next_split_text = next_split_text[end_index + 1 :]

                ind = -1 if is_setext_end else 0
                self.__fix_adjustments("\t", ind=ind)
            new_text_parts.append(next_split_text)

        if self.__seen_first_line_of_setext and next_split_end_whitespace is not None:
            self.__handle_text_split_end(
                next_split_end_whitespace, split_index, new_end_parts
            )
        else:
            self.__seen_first_line_of_setext = True
            if next_split_end_whitespace is not None:
                new_end_parts.append(next_split_end_whitespace)

    # pylint: enable=too-many-arguments

    def __handle_text(
        self,
        context: PluginScanContext,
        token: TextMarkdownToken,
        is_setext_end: bool = False,
    ) -> None:
        if (
            not self.__setext_start_token
            or (self.__any_leading_whitespace_detected and not context.in_fix_mode)
            or (not token.end_whitespace and not is_setext_end)
        ):
            self.__last_skipped_text_token = token
            return

        self.__last_skipped_text_token = None
        assert context.in_fix_mode or not self.__any_leading_whitespace_detected

        split_text = token.token_text.split(ParserHelper.newline_character)
        split_end_whitespace = None
        if token.end_whitespace is not None:
            split_end_whitespace = token.end_whitespace.split(
                ParserHelper.newline_character
            )
            assert len(split_text) == len(split_end_whitespace)

        new_text_parts: List[str] = []
        new_end_parts: List[str] = []
        for split_index in range(len(split_text)):
            self.__handle_text_split(
                context,
                split_text,
                split_end_whitespace,
                split_index,
                new_text_parts,
                new_end_parts,
                is_setext_end,
            )

        if context.in_fix_mode:
            new_end_parts_combined = "\n".join(new_end_parts)
            if (
                token.end_whitespace is not None
                and new_end_parts_combined != token.end_whitespace
            ):
                self.register_fix_token_request(
                    context,
                    token,
                    "next_token",
                    "end_whitespace",
                    new_end_parts_combined,
                )

            new_text_parts_combined = "\n".join(new_text_parts)
            if new_text_parts_combined != token.token_text:
                self.register_fix_token_request(
                    context,
                    token,
                    "next_token",
                    "token_text",
                    new_text_parts_combined,
                )

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        self.__container_manager.premanage_container_tokens(token)

        if token.is_atx_heading:
            atx_token = cast(AtxHeadingMarkdownToken, token)
            self.__handle_atx_heading(context, atx_token)
        elif token.is_setext_heading:
            setext_token = cast(SetextHeadingMarkdownToken, token)
            self.__handle_setext_heading(context, setext_token)
        elif token.is_text:
            text_token = cast(TextMarkdownToken, token)
            self.__handle_text(context, text_token)
        elif token.is_setext_heading_end:
            end_token = cast(EndMarkdownToken, token)
            self.__handle_setext_heading_end(context, end_token)

        if token.is_list_end:
            end_token = cast(EndMarkdownToken, token)
            list_start_token = end_token.start_markdown_token
            if list_start_token in self.__leading_spaces_split and context.in_fix_mode:
                assert list_start_token.is_list_start
                list_token = cast(ListStartMarkdownToken, list_start_token)
                new_leading_spaces = self.__leading_spaces_split[list_token]
                self.register_fix_token_request(
                    context,
                    list_start_token,
                    "next_token",
                    "leading_spaces",
                    "\n".join(new_leading_spaces),
                )

        self.__container_manager.manage_container_tokens(token)



"""
Module to help with the parsing of tabbified text inline elements.
"""

import logging
from typing import List, Optional

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.inline.inline_helper import InlineHelper
from pymarkdown.tokens.markdown_token import MarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


class InlineTabifiedTextBlockHelper:
    """
    Class to help with the parsing of tabbified text inline elements.
    """

    @staticmethod
    def handle_next_inline_character_tabified(
        source_text: str,
        tabified_text: str,
        newlines_encountered: int,
        start_index: int,
        next_index: int,
    ) -> Optional[str]:
        """
        Handle an inline character that is tabified.
        """
        adj_tabified_text, _ = InlineHelper.xdg(tabified_text, newlines_encountered)
        POGGER.debug("adj_tabified_text>:$:<", adj_tabified_text)

        _, line_start_index = InlineHelper.xdg(source_text, newlines_encountered)
        # POGGER.debug("adj_source_text>:$:<", adj_source_text)

        current_line_source_text = source_text[start_index:next_index]
        POGGER.debug("source_text>:$:<", source_text)
        POGGER.debug("current_line_source_text>:$:<", current_line_source_text)

        stop_character = source_text[next_index]
        if stop_character == "\n":
            return InlineTabifiedTextBlockHelper.__tabified_adjust(
                current_line_source_text, source_text, adj_tabified_text, start_index
            )

        POGGER.debug("stop_character>:$:<", stop_character)

        stop_character_in_tabified_index = InlineTabifiedTextBlockHelper.__handle_next_inline_character_tabified_find_stop(
            adj_tabified_text,
            stop_character,
            source_text,
            line_start_index,
            next_index,
        )

        (
            current_line_leading_space_index,
            current_line_leading_space,
        ) = ParserHelper.extract_spaces_verified(current_line_source_text, 0)
        POGGER.debug(
            "current_line_leading_space_index>:$:<, current_line_leading_space>:$:<",
            current_line_leading_space_index,
            current_line_leading_space,
        )

        return InlineTabifiedTextBlockHelper.__handle_next_inline_character_tabified_cleanup(
            current_line_source_text,
            current_line_leading_space_index,
            stop_character,
            adj_tabified_text,
            stop_character_in_tabified_index,
            current_line_leading_space,
        )

    @staticmethod
    def __tabified_adjust(
        current_line_source_text: str,
        source_text: str,
        adj_tabified_text: str,
        start_index: int,
    ) -> str:
        adj_current_line_source_text = current_line_source_text
        non_whitespace_index, ex_ws = ParserHelper.extract_spaces_verified(
            current_line_source_text, 0
        )
        was_last_character_newline = (
            True if start_index == 0 else source_text[start_index - 1] == "\n"
        )
        if ex_ws and was_last_character_newline:
            new_start = ParserHelper.create_replace_with_nothing_marker(ex_ws)
            adj_current_line_source_text = current_line_source_text[
                non_whitespace_index:
            ]
        else:
            new_start = ""

        ex_original_line, _ = TabHelper.find_detabify_string_ex(
            adj_tabified_text, adj_current_line_source_text
        )
        return new_start + ex_original_line

    @staticmethod
    def __handle_next_inline_character_tabified_find_stop(
        adj_tabified_text: str,
        stop_character: str,
        source_text: str,
        line_start_index: int,
        next_index: int,
    ) -> int:
        found_in_source_text_count = 0
        found_in_source_text_index = source_text.find(stop_character, line_start_index)
        POGGER.debug(
            "source_text[$]>:$:<",
            found_in_source_text_index,
            source_text[found_in_source_text_index:],
        )
        while found_in_source_text_index != next_index:
            found_in_source_text_count += 1
            found_in_source_text_index = source_text.find(
                stop_character, found_in_source_text_index + 1
            )
            POGGER.debug(
                "source_text[$]>:$:<",
                found_in_source_text_index,
                source_text[found_in_source_text_index:],
            )
        POGGER.debug("found_in_source_text_count>:$:<", found_in_source_text_count)
        POGGER.debug(
            "source_text[$]>:$:<",
            found_in_source_text_index,
            source_text[found_in_source_text_index:],
        )

        found_in_tabified_text_count = 0
        stop_character_in_tabified_index = adj_tabified_text.find(stop_character)
        POGGER.debug(
            "adj_tabified_text[$]>:$:<",
            stop_character_in_tabified_index,
            adj_tabified_text[stop_character_in_tabified_index:],
        )
        while found_in_tabified_text_count != found_in_source_text_count:
            found_in_tabified_text_count += 1
            stop_character_in_tabified_index = adj_tabified_text.find(
                stop_character, stop_character_in_tabified_index + 1
            )
            POGGER.debug(
                "adj_tabified_text[$]>:$:<",
                stop_character_in_tabified_index,
                adj_tabified_text[stop_character_in_tabified_index:],
            )
            assert (
                stop_character_in_tabified_index != -1
            ), "Stop character must be found."
        POGGER.debug("found_in_tabified_text_count>:$:<", found_in_tabified_text_count)
        POGGER.debug(
            "adj_tabified_text[$]>:$:<",
            stop_character_in_tabified_index,
            adj_tabified_text[stop_character_in_tabified_index:],
        )
        assert (
            adj_tabified_text[stop_character_in_tabified_index] == stop_character
        ), "Found character must be stop character."
        return stop_character_in_tabified_index

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_next_inline_character_tabified_cleanup(
        current_line_source_text: str,
        current_line_leading_space_index: int,
        stop_character: str,
        adj_tabified_text: str,
        stop_character_in_tabified_index: int,
        current_line_leading_space: str,
    ) -> str:
        (
            current_line_first_word_index,
            current_line_first_word,
        ) = ParserHelper.collect_until_one_of_characters_verified(
            current_line_source_text,
            current_line_leading_space_index,
            " \t" + stop_character,
        )
        POGGER.debug(
            "current_line_first_word_index>:$:<, current_line_first_word>:$:<",
            current_line_first_word_index,
            current_line_first_word,
        )

        if current_line_leading_space_index == current_line_first_word_index:
            tabified_start_index, _ = ParserHelper.extract_spaces_from_end(
                adj_tabified_text, stop_character_in_tabified_index
            )
        else:
            tabified_start_index = InlineHelper.pdff(
                current_line_source_text,
                current_line_first_word,
                current_line_leading_space_index,
                adj_tabified_text,
                stop_character_in_tabified_index,
                current_line_leading_space,
            )
        return adj_tabified_text[tabified_start_index:stop_character_in_tabified_index]

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def complete_inline_block_processing_tabified(
        source_text: str,
        start_index: int,
        tabified_text: str,
        newlines_encountered: int,
        inline_blocks: List[MarkdownToken],
        end_string: Optional[str],
    ) -> str:
        """
        Complete the processing on a tabified block
        """

        source_text_spaces_index, source_text_spaces = (
            ParserHelper.extract_spaces_verified(source_text, start_index)
        )
        POGGER.debug(
            "source_text_spaces_index=>:$:<, source_text_spaces=>:$:<",
            source_text_spaces_index,
            source_text_spaces,
        )
        (
            source_text_word_index,
            source_text_word,
        ) = ParserHelper.collect_until_one_of_characters_verified(
            source_text, source_text_spaces_index, " \t"
        )
        POGGER.debug(
            "source_text_word_index=>:$:<, source_text_word=>:$:<",
            source_text_word_index,
            source_text_word,
        )

        current_line_source_text, current_line_start_index = InlineHelper.xdf(
            source_text, newlines_encountered
        )
        POGGER.debug("current_line_source_text>:$:<", current_line_source_text)
        POGGER.debug("current_line_start_index>:$:<", current_line_start_index)

        find_word_count = 0
        word_index = current_line_source_text.find(source_text_word, 0)
        adj_word_index = word_index + current_line_start_index
        POGGER.debug(
            "[$]-->current_line_source_text[$:]>:$:<",
            find_word_count,
            word_index,
            current_line_source_text[word_index:],
        )
        POGGER.debug(
            "adj_word_index=$ != source_text_spaces_index=$",
            adj_word_index,
            source_text_spaces_index,
        )
        while adj_word_index != source_text_spaces_index:
            find_word_count += 1
            word_index = current_line_source_text.find(source_text_word, word_index + 1)
            adj_word_index = word_index + current_line_start_index
            POGGER.debug(
                "[$]-->current_line_source_text[$:]>:$:<",
                find_word_count,
                word_index,
                current_line_source_text[word_index:],
            )
            POGGER.debug(
                "adj_word_index=$ != source_text_spaces_index=$",
                adj_word_index,
                source_text_spaces_index,
            )
        POGGER.debug(
            "[$]-->current_line_source_text[$:]>:$:<",
            find_word_count,
            word_index,
            current_line_source_text[word_index:],
        )

        current_line_tabified_text, _ = InlineHelper.xdf(
            tabified_text, newlines_encountered
        )
        POGGER.debug("current_line_tabified_text>:$:<", current_line_tabified_text)

        if inline_blocks and inline_blocks[-1].is_inline_hard_break and not end_string:
            return InlineTabifiedTextBlockHelper.__tabified_adjust(
                current_line_source_text,
                source_text,
                current_line_tabified_text,
                current_line_start_index,
            )
        found_word_index = InlineHelper.calculate_word_index(
            current_line_tabified_text,
            source_text_word,
            find_word_count,
            source_text_spaces,
        )
        return current_line_tabified_text[found_word_index:]

    # pylint: enable=too-many-arguments, too-many-locals



"""
Module to implement a plugin that looks for inline HTML in the files.
"""

from typing import List, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.raw_html_markdown_token import RawHtmlMarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class RuleMd033(RulePlugin):
    """
    Class to implement a plugin that looks for inline HTML in the files.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__allowed_elements: List[str] = []
        self.__allow_first_image_element: bool = False
        self.__is_next_html_block_start: bool = False
        self.__is_first_element: bool = False
        self.__is_first_html_block: bool = False

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="no-inline-html",
            plugin_id="MD033",
            plugin_enabled_by_default=True,
            plugin_description="Inline HTML",
            plugin_version="0.6.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md033.md",
            plugin_configuration="allowed_elements, allow_first_image_element",
        )

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__allow_first_image_element = (
            self.plugin_configuration.get_boolean_property_with_default(
                "allow_first_image_element", True
            )
        )
        allowed_elements = self.plugin_configuration.get_string_property_with_default(
            "allowed_elements",
            "!--,![CDATA[,!DOCTYPE",
        )
        self.__allowed_elements = []
        if allowed_elements := allowed_elements.strip(" "):
            for next_element in allowed_elements.split(","):
                if next_element := next_element.strip(" "):
                    self.__allowed_elements.append(next_element)
                else:
                    raise ValueError(
                        "Elements in the comma-separated list cannot be empty."
                    )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem(
                "allow_first_image_element", self.__allow_first_image_element
            ),
            QueryConfigItem("allowed_elements", ",".join(self.__allowed_elements)),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__is_next_html_block_start = False
        self.__is_first_html_block = False
        self.__is_first_element = True

    def __look_for_html_start(
        self, context: PluginScanContext, token: MarkdownToken, tag_text: str
    ) -> None:
        # sourcery skip: extract-method, inline-immediately-returned-variable
        full_tag_text = tag_text.lower()
        if tag_text.startswith("/"):
            return
        if tag_text.startswith("![CDATA["):
            tag_text = "![CDATA["
        elif tag_text.startswith("!--"):
            tag_text = "!--"
        elif tag_text.startswith("!DOCTYPE"):
            tag_text = "!DOCTYPE"
        else:
            _, tag_text = ParserHelper.collect_until_one_of_characters_verified(
                tag_text, 0, " \n\t/>"
            )
        extra_data = f"Element: {tag_text}"

        if (
            self.__is_first_html_block
            and self.__allow_first_image_element
            and tag_text.lower() == "h1"
        ):
            is_first_image_element = full_tag_text.endswith("</h1>")
            if is_first_image_element:
                full_tag_text = full_tag_text[: -len("</h1>")]
                end_of_start_heading_index = full_tag_text.find(">")
                assert end_of_start_heading_index != -1
                full_tag_text = full_tag_text[end_of_start_heading_index + 1 :]
                end_of_image_index = full_tag_text.find(">")
                is_first_image_element = (
                    full_tag_text.startswith("<img")
                    and end_of_image_index == len(full_tag_text) - 1
                )
        else:
            is_first_image_element = False

        if not is_first_image_element and tag_text not in self.__allowed_elements:
            self.report_next_token_error(
                context, token, extra_error_information=extra_data
            )

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_inline_raw_html:
            self.__is_first_html_block = False
            raw_html_token = cast(RawHtmlMarkdownToken, token)
            self.__look_for_html_start(context, raw_html_token, raw_html_token.raw_tag)
        elif token.is_html_block:
            self.__is_next_html_block_start = True
            self.__is_first_html_block = self.__is_first_element
        elif token.is_text and self.__is_next_html_block_start:
            text_token = cast(TextMarkdownToken, token)
            self.__look_for_html_start(context, text_token, text_token.token_text[1:])
        else:
            self.__is_next_html_block_start = False

        self.__is_first_element = False



"""
Module to provide for an encapsulation of the ordered list start element.
"""

from typing import Optional

from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.list_start_markdown_token_helper import (
    ListStartMarkdownTokenHelper,
)
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_markdown.markdown_transform_context import (
    RegisterHtmlTransformHandlersProtocol,
)


class OrderedListStartMarkdownToken(ListStartMarkdownToken):
    """
    Class to provide for an encapsulation of the ordered list start element.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        list_start_sequence: str,
        list_start_content: str,
        indent_level: int,
        tabbed_adjust: int,
        extracted_whitespace: str,
        tabbed_whitespace_to_add: Optional[str],
        position_marker: PositionMarker,
    ) -> None:
        ListStartMarkdownToken.__init__(
            self,
            MarkdownToken._token_ordered_list_start,
            position_marker,
            list_start_sequence,
            list_start_content,
            indent_level,
            tabbed_adjust,
            extracted_whitespace,
            tabbed_whitespace_to_add,
        )

    # pylint: enable=too-many-arguments
    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_ordered_list_start

    # pylint: enable=protected-access

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            OrderedListStartMarkdownToken,
            ListStartMarkdownTokenHelper.handle_start_list_token,
            ListStartMarkdownTokenHelper.handle_end_list_token,
        )



"""
Module to provide for "-m pymarkdown" access to the module,
as if it was run from the console.
"""

from pymarkdown.main import PyMarkdownLint


def main() -> None:
    """
    Main entry point.  Exposed in this manner so that the setup
    entry_points configuration has something to execute.
    """
    PyMarkdownLint().main()


if __name__ == "__main__":
    main()



"""
Module to provide for an encapsulation of the indented code block element.
"""

from typing import Optional

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.tokens.leaf_markdown_token import LeafMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class IndentedCodeBlockMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the indented code block element.
    """

    def __init__(
        self, extracted_whitespace: str, line_number: int, column_number: int
    ) -> None:
        self.__indented_whitespace = ""
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_indented_code_block,
            extracted_whitespace,
            line_number=line_number,
            column_number=column_number,
            extracted_whitespace=extracted_whitespace,
            requires_end_token=True,
        )
        self.__compose_extra_data_field()

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_indented_code_block

    # pylint: enable=protected-access

    @property
    def indented_whitespace(self) -> str:
        """
        Returns any indented whitespace that comes before the text.
        """
        return self.__indented_whitespace

    def __compose_extra_data_field(self) -> None:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        self._set_extra_data(
            MarkdownToken.extra_data_separator.join(
                [self.extracted_whitespace, self.indented_whitespace]
            )
        )

    def add_indented_whitespace(self, indented_whitespace: str) -> None:
        """
        Add the indented whitespace that comes before the text.
        """
        self.__indented_whitespace = (
            f"{self.__indented_whitespace}{ParserHelper.newline_character}"
            + f"{indented_whitespace}"
        )
        self.__compose_extra_data_field()

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            IndentedCodeBlockMarkdownToken,
            IndentedCodeBlockMarkdownToken.__rehydrate_indented_code_block,
            IndentedCodeBlockMarkdownToken.__rehydrate_indented_code_block_end,
        )

    @staticmethod
    def __rehydrate_indented_code_block(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the indented code block from the token.
        """
        _ = previous_token

        context.block_stack.append(current_token)
        return ""

    @staticmethod
    def __rehydrate_indented_code_block_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the end of the indented code block from the token.
        """
        _ = (current_token, previous_token, next_token)

        del context.block_stack[-1]
        return ""

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            IndentedCodeBlockMarkdownToken,
            IndentedCodeBlockMarkdownToken.__handle_start_indented_code_block_token,
            IndentedCodeBlockMarkdownToken.__handle_end_indented_code_block_token,
        )

    @staticmethod
    def __handle_start_indented_code_block_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        token_parts = []
        if (
            not output_html
            and transform_state.transform_stack
            and transform_state.transform_stack[-1].endswith("<li>")
        ):
            token_parts.append(ParserHelper.newline_character)
        elif output_html and output_html[-1] != ParserHelper.newline_character:
            token_parts.extend([output_html, ParserHelper.newline_character])
        else:
            token_parts.append(output_html)
        transform_state.is_in_code_block, transform_state.is_in_fenced_code_block = (
            True,
            False,
        )
        token_parts.append("<pre><code>")
        return "".join(token_parts)

    @staticmethod
    def __handle_end_indented_code_block_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        transform_state.is_in_code_block = False
        return "".join(
            [
                output_html,
                ParserHelper.newline_character,
                "</code></pre>",
                ParserHelper.newline_character,
            ]
        )



"""
Module to keep track of a request made by a plugin rule.
"""

from dataclasses import dataclass
from typing import List

from pymarkdown.tokens.markdown_token import MarkdownToken


@dataclass
class ReplaceTokensRecord:
    """
    Record to keep track of a request made by a plugin rule.
    """

    plugin_id: str
    start_token: MarkdownToken
    end_token: MarkdownToken
    replacement_tokens: List[MarkdownToken]



"""
Module to provide a tokenization of a markdown-encoded string.
"""

from abc import ABC, abstractmethod, abstractproperty
from typing import List, Optional

from pymarkdown.general.parser_helper import ParserHelper


class SourceProvider(ABC):
    """
    Class to provide for an abstract definition of an instance that provides
    information about the input source.
    """

    # pylint: disable=deprecated-decorator
    @abstractproperty
    def is_at_end_of_file(self) -> bool:
        """
        Whether the provider has reached the end of the input.
        """

    # pylint: enable=deprecated-decorator

    @abstractmethod
    def get_next_line(self) -> Optional[str]:
        """
        Get the next line from the source provider.
        """


class InMemorySourceProvider(SourceProvider):
    """
    Class to provide for a source provider that is totally within memory.
    """

    def __init__(self, source_text: str) -> None:
        self.__next_line_tuple: List[str] = source_text.split(
            ParserHelper.newline_character, 1
        )

    @property
    def is_at_end_of_file(self) -> bool:
        """
        Whether the provider has reached the end of the input.
        """
        return not self.__next_line_tuple

    def get_next_line(self) -> Optional[str]:
        """
        Get the next line from the source provider.
        """
        token_to_use = None
        if not self.is_at_end_of_file:
            token_to_use = self.__next_line_tuple[0]
            if len(self.__next_line_tuple) == 2:
                self.__next_line_tuple = self.__next_line_tuple[1].split(
                    ParserHelper.newline_character, 1
                )
            else:
                assert self.__next_line_tuple
                self.__next_line_tuple = []
        return token_to_use


class FileSourceProvider(SourceProvider):
    """
    Class to provide for a source provider that is on media as a file.
    """

    def __init__(self, file_to_open: str) -> None:
        with open(file_to_open, encoding="utf-8") as file_to_parse:
            file_as_lines = file_to_parse.readlines()

        self.__read_lines, self.__read_index, did_line_end_in_newline = [], 0, True
        for next_line in file_as_lines:
            did_line_end_in_newline = next_line.endswith(ParserHelper.newline_character)
            if did_line_end_in_newline:
                next_line = next_line[:-1]
            self.__read_lines.append(next_line)

        self.__did_final_line_end_with_newline = did_line_end_in_newline
        if did_line_end_in_newline:
            self.__read_lines.append("")

    @property
    def did_final_line_end_with_newline(self) -> bool:
        """
        Indicate whether the final line of the file ended with a newline.
        """
        return self.__did_final_line_end_with_newline

    @property
    def is_at_end_of_file(self) -> bool:
        """
        Whether the provider has reached the end of the input.
        """
        return self.__read_index >= len(self.__read_lines)

    def get_next_line(self) -> Optional[str]:
        """
        Get the next line from the source provider.
        """
        token_to_use = None
        if not self.is_at_end_of_file:
            token_to_use = self.__read_lines[self.__read_index]
            self.__read_index += 1
        return token_to_use

    def reset_to_start(self) -> None:
        """
        Reset the provider to the start of the stream.
        """
        self.__read_index = 0



"""
Module to provide for the ability to parse the text for a link.
"""

import logging
import urllib
import urllib.parse
from typing import Dict, List, Optional, Tuple

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.inline.inline_backslash_helper import InlineBackslashHelper
from pymarkdown.inline.inline_helper import InlineHelper
from pymarkdown.inline.inline_request import InlineRequest
from pymarkdown.links.link_helper_properties import LinkHelperProperties
from pymarkdown.links.link_reference_titles import LinkReferenceTitles

POGGER = ParserLogger(logging.getLogger(__name__))


class LinkParseHelper:
    """
    Class to provide for the ability to parse the text for a link.
    """

    __link_definitions: Dict[str, LinkReferenceTitles] = {}

    __link_safe_characters = "/#:?=()*!$'+,;@"
    __special_link_destination_characters = "%&"

    __link_format_inline_start = "("
    __link_format_inline_end = ")"

    __link_title_single = "'"
    __link_title_double = '"'
    __link_title_parenthesis_open = "("
    __link_title_parenthesis_close = ")"

    __link_format_inline_end = ")"
    __link_format_reference_start = "["
    __link_format_reference_end = "]"

    link_label_start = "["
    link_label_end = "]"
    __link_label_is_definition_character = ":"
    __link_label_breaks = (
        f"{link_label_start}{link_label_end}{InlineBackslashHelper.backslash_character}"
    )

    __angle_link_start = "<"
    __angle_link_end = ">"
    __angle_link_destination_breaks = (
        f"{__angle_link_end}{InlineBackslashHelper.backslash_character}"
    )

    __non_angle_link_nest = "("
    __non_angle_link_unnest = ")"
    __non_angle_link_breaks = f"{Constants.ascii_control_characters}()\\"

    @staticmethod
    def initialize() -> None:
        """
        Initialize the inline subsystem.
        """
        LinkParseHelper.__link_definitions = {}

    @staticmethod
    def add_link_definition(link_name: str, link_value: LinkReferenceTitles) -> bool:
        """
        Add a link definition to the cache of links.
        """
        POGGER.debug(
            ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>$",
            LinkParseHelper.__link_definitions,
        )
        POGGER.debug(
            ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>$:$",
            link_name,
            link_value,
        )
        if did_add_definition := link_name not in LinkParseHelper.__link_definitions:
            LinkParseHelper.__link_definitions[link_name] = link_value
            POGGER.debug(">>added def>>$-->$", link_name, link_value)
        return did_add_definition

    @staticmethod
    def normalize_link_label(link_label: str) -> str:
        """
        Translate a link label into a normalized form to use for comparisons.
        """

        # Fold all whitespace characters (except for space) into a space character
        link_label = ParserHelper.replace_any_of(
            link_label,
            Constants.non_space_ascii_whitespace,
            ParserHelper.space_character,
        )

        # Fold multiple spaces into a single space character.
        # x = link_label.split(ParserHelper.space_character)
        split_label = [s for s in link_label.split(ParserHelper.space_character) if s]
        link_label = ParserHelper.space_character.join(split_label)

        # Fold the case of any characters to their lower equivalent.
        return link_label.casefold().strip(ParserHelper.space_character)

    @staticmethod
    def look_up_link(
        link_to_lookup: str, new_index: int, link_type: str
    ) -> Tuple[int, str, str]:
        """
        Look up a link to see if it is present.
        """

        POGGER.debug("pre>>$<<", link_to_lookup)
        link_to_lookup = ParserHelper.remove_all_from_text(link_to_lookup)
        POGGER.debug("mid(pre-norm)>>$<<", link_to_lookup)

        link_label = LinkParseHelper.normalize_link_label(link_to_lookup)
        POGGER.debug("post>>$<<", link_label)

        POGGER.debug("defs>>$<<", LinkParseHelper.__link_definitions)
        if not link_label or link_label not in LinkParseHelper.__link_definitions:
            update_index: int = -1
            inline_link: str = ""
            inline_title: str = ""
        else:
            POGGER.debug(link_type)
            link_titles = LinkParseHelper.__link_definitions[link_label]
            assert (
                link_titles.inline_link is not None
                and link_titles.inline_title is not None
            ), "Link and title must be defined by now."
            update_index, inline_link, inline_title = (
                new_index,
                link_titles.inline_link,
                link_titles.inline_title,
            )
        return update_index, inline_link, inline_title

    @staticmethod
    def __try_to_find_link_match(
        parser_properties: ParseBlockPassProperties,
        new_index: int,
        source_text: str,
        text_from_blocks: str,
        tabified_text: Optional[str],
    ) -> Tuple[str, bool, int, str, str, Optional[str]]:
        POGGER.debug("collapsed reference?")

        if tabified_text:
            text_to_scan = tabified_text
            tabified_new_index = LinkParseHelper.__translate_between_strings(
                source_text, tabified_text, new_index
            )
            POGGER.debug("tabified_new_index>:$:<", tabified_new_index)
            new_index = tabified_new_index
        else:
            text_to_scan = source_text

        after_open_index = new_index + 1
        if tried_full_reference_form := ParserHelper.is_character_at_index(
            text_to_scan, after_open_index, LinkParseHelper.__link_format_reference_end
        ):
            ex_label: Optional[str] = ""

            POGGER.debug("collapsed reference")
            POGGER.debug(">>$>>", text_from_blocks)
            update_index, inline_link, inline_title = LinkParseHelper.look_up_link(
                text_from_blocks,
                after_open_index + 1,
                "collapsed reference",
            )
            POGGER.debug("collapsed reference>update_index>$", update_index)
            label_type = Constants.link_type__collapsed
        else:
            (
                ex_label,
                label_type,
                inline_link,
                inline_title,
                update_index,
                tried_full_reference_form,
            ) = LinkParseHelper.__try_to_find_link_match_try_full(
                parser_properties, text_to_scan, after_open_index
            )

        if tabified_text and update_index != -1:
            # Both of the above functions consume the last character of the link.
            # Instead of guessing, we "rewind" the index by one character so that
            # we can have something to sync on that is not an end of line or whitespace.
            assert (
                tabified_text[update_index - 1]
                == LinkParseHelper.__link_format_reference_end
            ), "Values must be equal."
            untabified_update_index = LinkParseHelper.__translate_between_strings(
                tabified_text, source_text, update_index - 1
            )
            POGGER.debug("untabified_update_index>:$:<", untabified_update_index)
            update_index = untabified_update_index + 1

        return (
            label_type,
            tried_full_reference_form,
            update_index,
            inline_link,
            inline_title,
            ex_label,
        )

    @staticmethod
    def __try_to_find_link_match_try_full(
        parser_properties: ParseBlockPassProperties,
        text_to_scan: str,
        after_open_index: int,
    ) -> Tuple[Optional[str], str, str, str, int, bool]:
        POGGER.debug("full reference?")
        POGGER.debug(">>tried_full_reference_form>>$>", text_to_scan[after_open_index:])
        (
            tried_full_reference_form,
            after_label_index,
            ex_label,
        ) = LinkParseHelper.extract_link_label(
            parser_properties,
            text_to_scan,
            after_open_index,
            include_reference_colon=False,
        )
        POGGER.debug(
            ">>tried_full_reference_form>>$>after_label_index>$>ex_label>$>",
            tried_full_reference_form,
            after_label_index,
            ex_label,
        )
        if tried_full_reference_form:
            assert ex_label is not None, "If Full reference, the label is the link."
            label_type = Constants.link_type__full
            update_index, inline_link, inline_title = LinkParseHelper.look_up_link(
                ex_label, after_label_index, "full reference"
            )
        else:
            label_type, inline_link, inline_title, update_index = "", "", "", -1
        return (
            ex_label,
            label_type,
            inline_link,
            inline_title,
            update_index,
            tried_full_reference_form,
        )

    @staticmethod
    def __translate_between_strings(
        source_text: str, destination_text: str, next_index: int
    ) -> int:
        stop_character = source_text[next_index]

        POGGER.debug("source_text>:$:<", source_text)
        POGGER.debug("index>:$:< == >:$:<", next_index, stop_character)
        POGGER.debug("destination_text>:$:<", destination_text)

        found_in_source_text_count = 0
        found_in_source_text_index = source_text.find(stop_character)
        POGGER.debug(
            "source_text[$]>:$:<",
            found_in_source_text_index,
            source_text[found_in_source_text_index:],
        )
        while found_in_source_text_index != next_index:
            found_in_source_text_count += 1
            found_in_source_text_index = source_text.find(
                stop_character, found_in_source_text_index + 1
            )
            POGGER.debug(
                "source_text[$]>:$:<",
                found_in_source_text_index,
                source_text[found_in_source_text_index:],
            )
        POGGER.debug("found_in_source_text_count>:$:<", found_in_source_text_count)
        POGGER.debug(
            "source_text[$]>:$:<",
            found_in_source_text_index,
            source_text[found_in_source_text_index:],
        )

        found_in_destination_text_count = 0
        stop_character_in_destination_index = destination_text.find(stop_character)
        POGGER.debug(
            "stop_character_in_destination_index[$]>:$:<",
            stop_character_in_destination_index,
            destination_text[stop_character_in_destination_index:],
        )
        while found_in_destination_text_count != found_in_source_text_count:
            found_in_destination_text_count += 1
            stop_character_in_destination_index = destination_text.find(
                stop_character, stop_character_in_destination_index + 1
            )
            POGGER.debug(
                "adj_tabified_text[$]>:$:<",
                stop_character_in_destination_index,
                destination_text[stop_character_in_destination_index:],
            )
            assert (
                stop_character_in_destination_index != -1
            ), "Stop character must be found."
        POGGER.debug(
            "found_in_destination_text_count>:$:<", found_in_destination_text_count
        )
        POGGER.debug(
            "adj_tabified_text[$]>:$:<",
            stop_character_in_destination_index,
            destination_text[stop_character_in_destination_index:],
        )
        assert (
            destination_text[stop_character_in_destination_index] == stop_character
        ), "Stop index must refer to the expected stop character."
        return stop_character_in_destination_index

    @staticmethod
    def __parse_link_title(
        parser_properties: ParseBlockPassProperties, source_text: str, new_index: int
    ) -> Tuple[Optional[str], Optional[str], int, str]:
        """
        Parse an inline link's link title.
        """

        POGGER.debug("parse_link_title>>new_index>>$>>", source_text[new_index:])
        ex_title: Optional[str] = ""
        bounding_character = ""
        newer_index = new_index
        if ParserHelper.is_character_at_index(
            source_text, newer_index, LinkParseHelper.__link_title_single
        ):
            bounding_character = LinkParseHelper.__link_title_single
            newer_index, ex_title = InlineHelper.extract_bounded_string(
                parser_properties,
                source_text,
                newer_index + 1,
                LinkParseHelper.__link_title_single,
                None,
            )
        elif ParserHelper.is_character_at_index(
            source_text, newer_index, LinkParseHelper.__link_title_double
        ):
            bounding_character = LinkParseHelper.__link_title_double
            newer_index, ex_title = InlineHelper.extract_bounded_string(
                parser_properties,
                source_text,
                newer_index + 1,
                LinkParseHelper.__link_title_double,
                None,
            )
        elif ParserHelper.is_character_at_index(
            source_text, newer_index, LinkParseHelper.__link_title_parenthesis_open
        ):
            bounding_character = LinkParseHelper.__link_title_parenthesis_open
            newer_index, ex_title = InlineHelper.extract_bounded_string(
                parser_properties,
                source_text,
                newer_index + 1,
                LinkParseHelper.__link_title_parenthesis_close,
                LinkParseHelper.__link_title_parenthesis_open,
            )
        else:
            newer_index = -1
        POGGER.debug(
            "parse_link_title>>new_index>>$>>ex_link>>$>>",
            newer_index,
            ex_title,
        )
        pre_ex_title = ex_title
        if ex_title is not None:
            ex_title = InlineHelper.append_text(
                "",
                InlineBackslashHelper.handle_backslashes(parser_properties, ex_title),
                add_text_signature=False,
            )
        POGGER.debug("parse_link_title>>pre>>$>>", pre_ex_title)
        POGGER.debug("parse_link_title>>after>>$>>", ex_title)

        return ex_title, pre_ex_title, newer_index, bounding_character

    @staticmethod
    def extract_link_title(
        parser_properties: ParseBlockPassProperties,
        line_to_parse: str,
        new_index: int,
        is_blank_line: bool,
    ) -> Tuple[bool, int, Optional[str], Optional[str], Optional[str], Optional[str]]:
        """
        Extract the link reference definition's optional link title.
        """
        POGGER.debug("before ws>>$>", line_to_parse[new_index:])
        new_index, ex_ws = ParserHelper.extract_ascii_whitespace_verified(
            line_to_parse, new_index
        )
        POGGER.debug(
            "after ws>>$>ex_ws>$",
            line_to_parse[new_index:],
            ex_ws,
        )
        start_index, line_to_parse_size = new_index, len(line_to_parse)
        if new_index == line_to_parse_size and not is_blank_line:
            return False, new_index, None, None, None, None
        if ex_ws and new_index < line_to_parse_size:
            (
                inline_title,
                pre_inline_title,
                new_index,
                _,
            ) = LinkParseHelper.__parse_link_title(
                parser_properties, line_to_parse, new_index
            )
            if new_index == -1 or inline_title is None:
                return False, new_index, None, None, None, None
        else:
            inline_title, pre_inline_title = "", ""
        return (
            True,
            new_index,
            inline_title,
            pre_inline_title,
            ex_ws,
            line_to_parse[start_index:new_index],
        )

    @staticmethod
    def __encode_link_destination(link_to_encode: str) -> str:
        percent_index, before_data = (
            ParserHelper.collect_until_one_of_characters_verified(
                link_to_encode, 0, LinkParseHelper.__special_link_destination_characters
            )
        )
        el_parts, link_to_encode_size = [
            urllib.parse.quote(before_data, safe=LinkParseHelper.__link_safe_characters)
        ], len(link_to_encode)
        while percent_index < link_to_encode_size:
            special_character = link_to_encode[percent_index]
            percent_index += 1
            if special_character == "%":
                hex_guess_characters = link_to_encode[percent_index : percent_index + 2]
                if len(hex_guess_characters) == 2:
                    try:
                        int(hex_guess_characters, 16)
                        el_parts.extend(["%", hex_guess_characters])
                        percent_index += 2
                    except ValueError:
                        el_parts.append("%25")
                else:
                    el_parts.append("%25")
            else:
                assert (
                    special_character == "&"
                ), "Special character is either % (see if) or &."
                el_parts.append("&amp;")

            percent_index, before_data = (
                ParserHelper.collect_until_one_of_characters_verified(
                    link_to_encode,
                    percent_index,
                    LinkParseHelper.__special_link_destination_characters,
                )
            )
            el_parts.append(
                urllib.parse.quote(
                    before_data, safe=LinkParseHelper.__link_safe_characters
                )
            )

        return "".join(el_parts)

    @staticmethod
    def __parse_link_destination(
        parser_properties: ParseBlockPassProperties, source_text: str, new_index: int
    ) -> Tuple[Optional[str], Optional[str], int, Optional[str], Optional[bool]]:
        """
        Parse an inline link's link destination.
        """

        POGGER.debug("parse_link_destination>>new_index>>$>>", source_text[new_index:])
        start_index = new_index
        ex_link: Optional[str] = ""
        if did_use_angle_start := ParserHelper.is_character_at_index(
            source_text, new_index, LinkParseHelper.__angle_link_start
        ):
            POGGER.debug(
                ">parse_angle_link_destination>new_index>$>$",
                new_index,
                source_text[new_index:],
            )
            new_index, ex_link = LinkParseHelper.__parse_angle_link_destination(
                parser_properties, source_text, new_index
            )
            POGGER.debug(
                ">parse_angle_link_destination>new_index>$>ex_link>$>",
                new_index,
                ex_link,
            )
        else:
            POGGER.debug(
                ">parse_non_angle_link_destination>new_index>$>$",
                new_index,
                source_text[new_index:],
            )
            newer_index, ex_link = LinkParseHelper.__parse_non_angle_link_destination(
                parser_properties, source_text, new_index
            )
            new_index = newer_index
            POGGER.debug(
                ">parse_non_angle_link_destination>new_index>$>ex_link>$>",
                new_index,
                ex_link,
            )
            if not ex_link:
                return None, None, -1, None, None

        if new_index != -1 and ParserHelper.newline_character in ex_link:
            return None, None, -1, None, None
        POGGER.debug(
            "handle_backslashes>>new_index>>$>>ex_link>>$>>",
            new_index,
            ex_link,
        )

        pre_handle_link = ex_link
        if new_index != -1 and ex_link:
            ex_link = InlineBackslashHelper.handle_backslashes(
                parser_properties, ex_link
            )
        POGGER.debug(
            "urllib.parse.quote>>ex_link>>$>>",
            ex_link,
        )

        ex_link = LinkParseHelper.__encode_link_destination(ex_link)
        POGGER.debug(
            "parse_link_destination>>new_index>>$>>ex_link>>$>>",
            new_index,
            ex_link,
        )
        return (
            ex_link,
            pre_handle_link,
            new_index,
            source_text[start_index:new_index],
            did_use_angle_start,
        )

    @staticmethod
    def __parse_non_angle_link_destination(
        parser_properties: ParseBlockPassProperties, source_text: str, new_index: int
    ) -> Tuple[int, Optional[str]]:
        """
        Parse a link destination that is not included in angle brackets.
        """

        destination_parts: List[str] = []
        keep_collecting, nesting_level = True, 0
        newer_index = new_index
        while keep_collecting:
            POGGER.debug(
                "collected_destination>>$<<source_text<<$>>nesting_level>>$>>",
                destination_parts,
                source_text[newer_index:],
                nesting_level,
            )
            newer_index, before_part = (
                ParserHelper.collect_until_one_of_characters_verified(
                    source_text, newer_index, LinkParseHelper.__non_angle_link_breaks
                )
            )
            destination_parts.append(before_part)
            POGGER.debug(">>>>>>$<<<<<", source_text[newer_index:])
            if ParserHelper.is_character_at_index(
                source_text, newer_index, InlineBackslashHelper.backslash_character
            ):
                POGGER.debug("backslash")
                old_new_index = newer_index
                inline_request = InlineRequest(source_text, newer_index)
                inline_response = InlineBackslashHelper.handle_inline_backslash(
                    parser_properties, inline_request
                )
                assert (
                    inline_response.new_index is not None
                ), "Index must be defined by now."
                newer_index = inline_response.new_index
                destination_parts.append(source_text[old_new_index:newer_index])
            elif ParserHelper.is_character_at_index(
                source_text, newer_index, LinkParseHelper.__non_angle_link_nest
            ):
                POGGER.debug("+1")
                nesting_level += 1
                destination_parts.append(LinkParseHelper.__non_angle_link_nest)
                newer_index += 1
            elif ParserHelper.is_character_at_index(
                source_text, newer_index, LinkParseHelper.__non_angle_link_unnest
            ):
                POGGER.debug("-1")
                if keep_collecting := bool(nesting_level):
                    destination_parts.append(LinkParseHelper.__non_angle_link_unnest)
                    newer_index += 1
                    nesting_level -= 1
            else:
                keep_collecting = False

        if nesting_level:
            return -1, None
        return newer_index, "".join(destination_parts)

    @staticmethod
    def __parse_angle_link_destination(
        parser_properties: ParseBlockPassProperties, source_text: str, new_index: int
    ) -> Tuple[int, str]:
        """
        Parse a link destination that is included in angle brackets.
        """

        destination_parts: List[str] = []
        newer_index = new_index + 1
        while True:
            newer_index, ert_new = (
                ParserHelper.collect_until_one_of_characters_verified(
                    source_text,
                    newer_index,
                    LinkParseHelper.__angle_link_destination_breaks,
                )
            )
            destination_parts.append(ert_new)

            if not ParserHelper.is_character_at_index(
                source_text, newer_index, InlineBackslashHelper.backslash_character
            ):
                break

            old_new_index = newer_index
            inline_request = InlineRequest(source_text, newer_index)
            inline_response = InlineBackslashHelper.handle_inline_backslash(
                parser_properties, inline_request
            )
            assert (
                inline_response.new_index is not None
            ), "Index must be defined by now."
            newer_index = inline_response.new_index
            destination_parts.append(source_text[old_new_index:newer_index])
        if ParserHelper.is_character_at_index(
            source_text, newer_index, LinkParseHelper.__angle_link_end
        ):
            newer_index += 1
        else:
            newer_index = -1
            destination_parts.clear()
        return newer_index, "".join(destination_parts)

    @staticmethod
    def __parse_inline_link_properties(
        parser_properties: ParseBlockPassProperties,
        source_text: str,
        new_index: int,
        lhp: LinkHelperProperties,
    ) -> int:
        lhp.inline_title = ""
        lhp.pre_inline_title = ""
        lhp.bounding_character = ""
        lhp.before_title_whitespace = ""
        lhp.after_title_whitespace = ""
        temp_bool = None
        POGGER.debug(">>search for link destination")
        (
            lhp.inline_link,
            lhp.pre_inline_link,
            newer_index,
            _,
            temp_bool,
        ) = LinkParseHelper.__parse_link_destination(
            parser_properties, source_text, new_index
        )
        lhp.did_use_angle_start = temp_bool
        POGGER.debug(
            ">>link destination>>$>>$>>",
            source_text,
            newer_index,
        )
        if newer_index != -1:
            POGGER.debug(
                "before ws>>$<",
                source_text[newer_index:],
            )
            (
                newer_index,
                lhp.before_title_whitespace,
            ) = ParserHelper.extract_ascii_whitespace_verified(source_text, newer_index)
            POGGER.debug(
                "after ws>>$>",
                source_text[newer_index:],
            )
            if ParserHelper.is_character_at_index_not(
                source_text, newer_index, LinkParseHelper.__link_format_inline_end
            ):
                (
                    lhp.inline_title,
                    lhp.pre_inline_title,
                    newer_index,
                    lhp.bounding_character,
                ) = LinkParseHelper.__parse_link_title(
                    parser_properties, source_text, newer_index
                )
        if newer_index != -1:
            (
                newer_index,
                lhp.after_title_whitespace,
            ) = ParserHelper.extract_ascii_whitespace_verified(source_text, newer_index)
        return newer_index

    @staticmethod
    def extract_link_label(
        parser_properties: ParseBlockPassProperties,
        line_to_parse: str,
        new_index: int,
        include_reference_colon: bool = True,
    ) -> Tuple[bool, int, Optional[str]]:
        """
        Extract the link reference definition's link label.
        """
        label_parts: List[str] = []

        while True:
            new_index, ert_new = ParserHelper.collect_until_one_of_characters_verified(
                line_to_parse, new_index, LinkParseHelper.__link_label_breaks
            )
            label_parts.append(ert_new)
            if ParserHelper.is_character_at_index(
                line_to_parse, new_index, InlineBackslashHelper.backslash_character
            ):
                old_new_index = new_index
                inline_response = InlineBackslashHelper.handle_inline_backslash(
                    parser_properties, InlineRequest(line_to_parse, new_index)
                )
                assert (
                    inline_response.new_index is not None
                ), "TODO: can these be wrapped?"
                new_index = inline_response.new_index
                label_parts.append(line_to_parse[old_new_index:new_index])
            elif ParserHelper.is_character_at_index(
                line_to_parse, new_index, LinkParseHelper.link_label_start
            ):
                POGGER.debug(">> unescaped [, bailing")
                return False, -1, None
            else:
                break  # pragma: no cover

        POGGER.debug("look for ]>>$<<", line_to_parse[new_index:])
        if not ParserHelper.is_character_at_index(
            line_to_parse, new_index, LinkParseHelper.link_label_end
        ):
            POGGER.debug(">> no end ], bailing")
            return False, new_index, None
        new_index += 1

        if include_reference_colon:
            POGGER.debug(
                "look for :>>$<<",
                line_to_parse[new_index:],
            )
            if not ParserHelper.is_character_at_index(
                line_to_parse,
                new_index,
                LinkParseHelper.__link_label_is_definition_character,
            ):
                POGGER.debug(">> no :, bailing")
                return False, -1, None
            new_index += 1

        return True, new_index, "".join(label_parts)

    @staticmethod
    def extract_link_destination(
        parser_properties: ParseBlockPassProperties,
        line_to_parse: str,
        start_index: int,
        is_blank_line: bool,
    ) -> Tuple[bool, int, Optional[str], Optional[str], Optional[str], Optional[str]]:
        """
        Extract the link reference definition's link destination.
        """
        (
            after_whitespace_index,
            prefix_whitespace,
        ) = ParserHelper.collect_while_one_of_characters_verified(
            line_to_parse, start_index, Constants.ascii_whitespace
        )
        if after_whitespace_index == len(line_to_parse) and not is_blank_line:
            return False, after_whitespace_index, None, None, None, None
        POGGER.debug(
            "Pre-LD>>$<<", ParserHelper.make_whitespace_visible(prefix_whitespace)
        )
        POGGER.debug("LD>>$<<", line_to_parse[after_whitespace_index:])
        (
            inline_link,
            pre_inline_link,
            after_whitespace_index,
            inline_raw_link,
            _,
        ) = LinkParseHelper.__parse_link_destination(
            parser_properties, line_to_parse, after_whitespace_index
        )
        if after_whitespace_index == -1:
            return False, -1, None, None, None, None
        return (
            True,
            after_whitespace_index,
            inline_link,
            pre_inline_link,
            prefix_whitespace,
            inline_raw_link,
        )

    @staticmethod
    def __process_inline_link_body(
        parser_properties: ParseBlockPassProperties,
        source_text: str,
        new_index: int,
        tabified_text: Optional[str],
        lhp: LinkHelperProperties,
    ) -> int:
        """
        Given that an inline link has been identified, process it's body.
        """

        POGGER.debug("process_inline_link_body>:$:<", source_text[new_index:])
        POGGER.debug("source_text>:$:<", source_text)
        POGGER.debug("tabified_text>:$:<", tabified_text)

        if tabified_text:
            text_to_scan = tabified_text
            tabified_new_index = LinkParseHelper.__translate_between_strings(
                source_text, tabified_text, new_index
            )
            POGGER.debug("tabified_new_index>:$:<", tabified_new_index)
            new_index = tabified_new_index
        else:
            text_to_scan = source_text
        new_index += 1

        newer_index, lhp.before_link_whitespace = (
            ParserHelper.extract_ascii_whitespace_verified(text_to_scan, new_index)
        )

        POGGER.debug(
            "newer_index>>$>>text_to_scan[]>>$>",
            newer_index,
            text_to_scan[newer_index:],
        )
        if not ParserHelper.is_character_at_index(
            text_to_scan, newer_index, LinkParseHelper.__link_format_inline_end
        ):
            newer_index = LinkParseHelper.__parse_inline_link_properties(
                parser_properties, text_to_scan, newer_index, lhp
            )
        else:
            (
                lhp.inline_link,
                lhp.pre_inline_link,
                lhp.inline_title,
                lhp.pre_inline_title,
                lhp.did_use_angle_start,
                lhp.bounding_character,
                lhp.before_title_whitespace,
                lhp.after_title_whitespace,
            ) = ("", "", "", "", False, "", "", "")
        POGGER.debug(
            "inline_link>>$>>inline_title>>$>newer_index>$>",
            lhp.inline_link,
            lhp.inline_title,
            newer_index,
        )
        (
            newer_index,
            lhp.did_use_angle_start,
        ) = LinkParseHelper.__process_inline_link_body_final(
            newer_index, source_text, tabified_text, lhp.did_use_angle_start
        )
        POGGER.debug(
            "process_inline_link_body>>inline_link>>$>>inline_title>>$>new_index>$>",
            lhp.inline_link,
            lhp.inline_title,
            newer_index,
        )
        return newer_index

    @staticmethod
    def __process_inline_link_body_final(
        newer_index: int,
        source_text: str,
        tabified_text: Optional[str],
        did_use_angle_start: Optional[bool],
    ) -> Tuple[int, bool]:
        if newer_index != -1:
            assert (
                did_use_angle_start is not None
            ), "If index is not -1, did_use_angle_start must be defined."
            if tabified_text:
                untabified_newer_index = LinkParseHelper.__translate_between_strings(
                    tabified_text, source_text, newer_index
                )
                POGGER.debug("untabified_newer_index>:$:<", untabified_newer_index)
                newer_index = untabified_newer_index

            if ParserHelper.is_character_at_index(
                source_text, newer_index, LinkParseHelper.__link_format_inline_end
            ):
                newer_index += 1
            else:
                newer_index = -1
        else:
            did_use_angle_start = False
        return newer_index, did_use_angle_start

    @staticmethod
    def look_for_link_formats(
        parser_properties: ParseBlockPassProperties,
        source_text: str,
        new_index: int,
        text_from_blocks: str,
        tabified_text: Optional[str],
    ) -> Tuple[
        int,
        bool,
        LinkHelperProperties,
    ]:
        """
        Look for links in the various formats.
        """
        (
            update_index,
            tried_full_reference_form,
        ) = (-1, False)
        lhp: LinkHelperProperties = LinkHelperProperties()
        lhp.did_use_angle_start = False
        lhp.bounding_character = ""
        lhp.label_type = ""
        lhp.inline_link = ""
        lhp.pre_inline_link = ""
        lhp.inline_title = ""
        lhp.pre_inline_title = ""
        lhp.ex_label = ""
        lhp.before_link_whitespace = ""
        lhp.before_title_whitespace = ""
        lhp.after_title_whitespace = ""

        if ParserHelper.is_character_at_index(
            source_text, new_index, LinkParseHelper.__link_format_inline_start
        ):
            POGGER.debug("inline reference? >>$>>", new_index)
            update_index = LinkParseHelper.__process_inline_link_body(
                parser_properties, source_text, new_index, tabified_text, lhp
            )
            lhp.label_type = Constants.link_type__inline
        elif ParserHelper.is_character_at_index(
            source_text, new_index, LinkParseHelper.__link_format_reference_start
        ):
            (
                lhp.label_type,
                tried_full_reference_form,
                update_index,
                lhp.inline_link,
                lhp.inline_title,
                lhp.ex_label,
            ) = LinkParseHelper.__try_to_find_link_match(
                parser_properties,
                new_index,
                source_text,
                text_from_blocks,
                tabified_text,
            )
        POGGER.debug("__look_for_link_formats>>update_index>>$>>", update_index)
        return (
            update_index,
            tried_full_reference_form,
            lhp,
        )
